"""Class to hold Nexus PVT properties."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_BLACKOIL_PRIMARY_KEYWORDS, PVT_TYPE_KEYWORDS, PVT_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_EOS_METHODS, PVT_EOSOPTIONS_PRIMARY_WORDS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_EOSOPTIONS_PRIMARY_KEYS_INT, PVT_TABLE_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_TABLES_WITH_ENDWORDS, PVT_TABLES_WITHOUT_ENDWORDS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_EOSOPTIONS_PRIMARY_KEYS_FLOAT, PVT_EOSOPTIONS_TRANS_KEYS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_EOSOPTIONS_TRANS_TEST_KEYS, PVT_EOSOPTIONS_PHASEID_KEYS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_EOSOPTIONS_TERTIARY_KEYS, PVT_ALL_TABLE_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.pvt_keywords import PVT_UNSAT_TABLE_INDICES
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits
from ResSimpy.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import PVTUnits

from ResSimpy.Utils.factory_methods import get_empty_dict_union, get_empty_list_str, get_empty_eosopt_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusPVTMethod(DynamicProperty):
    """Class to hold Nexus PVT properties.

    Attributes:
        file (NexusFile): Nexus PVT file object
        input_number (int): PVT method number in Nexus fcs file
        pvt_type (Optional[str]): Type of PVT method, e.g., BLACKOIL, GASWATER or EOS. Defaults to None
        eos_nhc (Optional[int]): Number of hydrocarbon components. Defaults to None
        eos_temp (Optional[float]): Default temperature for EOS method. Defaults to None
        eos_components (Optional[list[str]]): Specifies component names
        eos_options (dict[str, Union[str, int, float, pd.DataFrame, list[str], dict[str, float],
            tuple[str, dict[str, float]], dict[str, pd.DataFrame]]]): Dictionary containing various EOS options
            as specified in the PVT file. Defaults to empty dictionary.
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific PVT method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile
    pvt_type: Optional[str] = None
    eos_nhc: Optional[int] = None  # Number of hydrocarbon components
    eos_temp: Optional[float] = None  # Default temperature for EOS method
    eos_components: Optional[list[str]] = field(default_factory=get_empty_list_str)
    eos_options: dict[str, Union[
        str, int, float, pd.DataFrame, list[str], dict[str, float], tuple[str, dict[str, float]], dict[
            str, pd.DataFrame]]] \
        = field(default_factory=get_empty_eosopt_dict_union)
    properties: dict[str, Union[str, int, float, Enum, list[str],
                                pd.DataFrame, dict[str, Union[float, pd.DataFrame]]]] \
        = field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 pvt_type: Optional[str] = None,
                 eos_nhc: Optional[int] = None, eos_temp: Optional[float] = None,
                 eos_components: Optional[list[str]] = None,
                 eos_options: Optional[dict[str, Union[str, int, float, pd.DataFrame, list[str], dict[str, float],
                                       tuple[str, dict[str, float]], dict[str, pd.DataFrame]]]] = None,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        if pvt_type is not None:
            self.pvt_type = pvt_type
        if eos_nhc is not None:
            self.eos_nhc = eos_nhc
        if eos_temp is not None:
            self.eos_temp = eos_temp
        if eos_components is not None:
            self.eos_components = eos_components
        else:
            self.eos_components = []
        if eos_options is not None:
            self.eos_options = eos_options
        else:
            self.eos_options = {}
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'PRES': ('pressure', float),
            'TEMP': ('temperature', float),
            'DP': ('delta_pressure', float),
            'DENOIL': ('oil_density', float),
            'DENGAS': ('gas_density', float),
            'API': ('oil_api_gravity', float),
            'SPECG': ('gas_specific_gravity', float),
            'MWOR': ('molecular_weight_of_residual_oil', float),
            'VO': ('oil_viscosity', float),
            'VG': ('gas_viscosity', float),
            'BO': ('oil_formation_volume_factor', float),
            'BG': ('gas_formation_volume_factor', float),
            'RS': ('solution_gas_oil_ratio', float),
            'RV': ('solution_oil_gas_ratio', float),
            'PSAT': ('saturation_pressure', float),
            'GOR': ('gas_oil_ratio', float),
            'OGR': ('oil_gas_ratio', float),
            'CV': ('gas_heat_capacity_at_constant_volume', float),
            'CP': ('gas_heat_capacity_at_constant_pressure', float),
            'BOFAC': ('oil_fvf_over_oil_fvf_at_saturation_pressure', float),
            'VOFAC': ('oil_viscosity_over_oil_viscosity_at_saturation_pressure', float),
            'BGFAC': ('gas_fvf_over_gas_fvf_at_saturation_pressure', float),
            'VGFAC': ('gas_viscosity_over_gas_viscosity_at_saturation_pressure', float),
            'CVFAC': ('gas_heat_capacity_at_constant_volume_over_cv_at_saturation_pressure', float),
            'CPFAC': ('gas_heat_capacity_at_constant_pressure_over_cp_at_saturation_pressure', float),
            'PC': ('critical_pressure', float),
            'TC': ('critical_temperature', float),
            'VC': ('critical_volume', float),
        }
        return keywords

    @property
    def units(self) -> PVTUnits:
        """Returns the attribute to unit map for the PVT method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return PVTUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with PVT data in Nexus file format."""
        printable_str = ''
        pvt_dict = self.properties
        # Print description if present
        if 'DESC' in pvt_dict.keys() and isinstance(pvt_dict['DESC'], list):
            for desc_line in pvt_dict['DESC']:
                printable_str += 'DESC ' + desc_line + '\n'
        # Print PVT type and associated properties
        printable_str += f'{self.pvt_type}'
        if self.eos_nhc is not None:
            printable_str += f' NHC {self.eos_nhc}'
        for pvt_key in PVT_BLACKOIL_PRIMARY_KEYWORDS:
            if pvt_key in pvt_dict.keys():
                printable_str += f' {pvt_key} {pvt_dict[pvt_key]}'
        printable_str += '\n'
        if 'DRYGAS_MFP' in pvt_dict.keys():
            printable_str += 'DRYGAS_MFP\n'
        # Print EOS-specific required parameters, if present
        if self.eos_components is not None:
            if len(self.eos_components) > 0:
                printable_str += f"COMPONENTS {' '.join(self.eos_components)}\n"
        if self.eos_temp is not None:
            printable_str += f'TEMP {self.eos_temp}\n'
        for key, value in pvt_dict.items():
            if isinstance(value, pd.DataFrame):
                printable_str += f'{key}\n'
                if key == 'IGS_CP':
                    printable_str += value.to_string(na_rep='', index=False, header=False) + '\n'
                else:
                    printable_str += value.to_string(na_rep='', index=False) + '\n'
                if key in PVT_TABLES_WITH_ENDWORDS:
                    printable_str += 'END'+key+'\n'
                printable_str += '\n'
            elif isinstance(value, dict):
                for subkey in value.keys():
                    printable_str += f"{key.replace('_',' ')} {subkey}\n"
                    df = value[subkey]
                    if isinstance(df, pd.DataFrame):
                        printable_str += df.to_string(na_rep='', index=False) + '\n'
                    printable_str += '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif key not in [*PVT_BLACKOIL_PRIMARY_KEYWORDS, 'NHC', 'DESC', 'DRYGAS_MFP']:
                if value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'
        if len(self.eos_options.keys()) > 0:
            pvt_dict = self.eos_options
            printable_str += 'EOSOPTIONS'
            if 'EOS_METHOD' in pvt_dict.keys():
                printable_str += f" {pvt_dict['EOS_METHOD']}"
            printable_str += '\n'
            for key, value in pvt_dict.items():
                if key == 'EOS_OPT_PRIMARY_LIST' and isinstance(value, list):
                    for token in value:
                        printable_str += f'    {token}\n'
                elif isinstance(value, tuple):
                    printable_str += f'    {key} {value[0]}'
                    val_dict: dict[str, float] = value[1]
                    for val_key, val_val in val_dict.items():
                        printable_str += f' {val_key} {val_val}'
                    printable_str += '\n'
                elif isinstance(value, dict):
                    printable_str += f'    {key}'
                    for subkey, subvalue in value.items():
                        printable_str += f' {subkey} {subvalue}'
                    printable_str += '\n'
                elif key not in ['EOS_METHOD']:
                    printable_str += f'    {key} {value}\n'
        return printable_str

    def __populate_eos_opts_to_tertiary_keys(self, primary_key: str, primary_key_default_val: str, single_line: str,
                                             line_list: list[str], list_of_secondary_keys: list[str]):
        """Utility function to populate complex EOS options structures, from primary to tertiary keyword level.
        Applies to TRANSITION, TRANS_TEST and PHASEID Nexus EOS options.

        Args:
            primary_key (str): primary keyword, e.g., TRANSITION or PHASEID
            primary_key_default_val (str): default secondary keyword, or primary key value, e.g., TEST
            single_line (str): single line as read from input PVT file
            line_list (list[str]): list of strings that comprise input PVT file
            list_of_secondary_keys (list[str]): list of secondary keywords associated with the given primary keyword
        """
        if nfo.check_token(primary_key, single_line):
            self.eos_options[primary_key] = primary_key_default_val  # Set default value
            if nfo.get_expected_token_value(primary_key, single_line, line_list) in list_of_secondary_keys:
                self.eos_options[primary_key] = nfo.get_expected_token_value(primary_key, single_line, line_list)
        if [i for i in single_line.split() if i in list_of_secondary_keys]:
            for secondary_key in list_of_secondary_keys:
                if nfo.check_token(secondary_key, single_line):
                    self.eos_options[primary_key] = secondary_key
                    for tertiary_key in PVT_EOSOPTIONS_TERTIARY_KEYS:
                        if nfo.check_token(tertiary_key, single_line):
                            if isinstance(self.eos_options[primary_key], str):  # Convert to tuple
                                self.eos_options[primary_key] = (secondary_key, {})

                            secondary_eos_option = self.eos_options[primary_key]
                            if not isinstance(secondary_eos_option, tuple) or \
                                    not isinstance(secondary_eos_option[1], dict):
                                raise ValueError(f"EOS secondary key invalid: {secondary_key}")
                            secondary_eos_option[1][tertiary_key] = float(
                                nfo.get_expected_token_value(tertiary_key, single_line, line_list))

    def __find_pvt_table_starting_index(self, table_key: str, single_line: str, line_list: list[str],
                                        table_indices: dict[str, list[int]],
                                        table_indices_dict: dict[str, dict[str, list[int]]],
                                        table_flag: dict[str, bool], l_index: int,
                                        unsat_obj: dict[str, list[str]] | None = None) -> Optional[int]:
        """Utility function to find the starting line index of a specified PVT table.

        Args:
            table_key (str): specified PVT table name or undersaturated index, such as, PSAT or RSSAT or PRES
            single_line (str): single line as read from input PVT file
            line_list (list[str]): list of strings that comprise input PVT file
            table_indices ([dict[str, list[int]]): dictionary to store the
                starting and ending line index of tables
            table_indices_dict (dict[str, dict[str, list[int]]]): dictionary to store the
                starting and ending line index of tables, for undersaturated tables
            table_flag (bool): flag to tell if a table is currently being read (True) or not (False)
            l_index (int): current line index
            unsat_obj (dict[str, list[str]]): track saturation pressures from which undersaturated branches emanate

        Raises:
            ValueError: If a property table key does not have a numerical value

        Returns:
            int: Updated line index
        """
        if unsat_obj is None:
            unsat_obj = {}
        if table_key not in PVT_UNSAT_TABLE_INDICES:  # All tables except undersaturated tables
            if nfo.check_token(table_key, single_line):
                table_indices[table_key] = [l_index + 1, len(line_list)]
                table_flag[table_key] = True
                return l_index + 1
        else:  # Handle undersaturated tables
            if table_key == 'PRES':
                table_name = 'UNSATGAS'
                full_table_name = table_name + '_PRES'
            else:
                table_name = 'UNSATOIL'
                full_table_name = table_name + '_' + table_key
            if nfo.check_token(table_name, single_line) and nfo.check_token(table_key, single_line):
                if nfo.get_token_value(table_key, single_line, line_list) is None:
                    raise ValueError(f'Property {table_key} does not have a numerical value.')
                unsat_obj[table_key].append(nfo.get_expected_token_value(table_key, single_line, line_list))
                if full_table_name in table_indices_dict.keys():
                    table_indices_dict[full_table_name][unsat_obj[table_key][-1]] = [l_index + 1, len(line_list)]
                else:
                    table_indices_dict[full_table_name] = {unsat_obj[table_key][-1]: [l_index + 1, len(line_list)]}
                table_flag[table_name] = True
                return l_index + 1
        return None

    def __find_pvt_table_ending_index(self, table_key: str, single_line: str,
                                      table_indices: dict[str, list[int]],
                                      table_indices_dict: dict[str, dict[str, list[int]]],
                                      l_index: int, table_flag: dict[str, bool],
                                      table_has_endkeyword: bool, reading_flag: bool,
                                      unsat_obj: dict[str, list[str]] = {}) -> bool:
        """Utility function to find the ending line index of a specified PVT table.

        Args:
            table_key (str): specified PVT table name or undersaturated index, such as, PSAT or RSSAT or PRES
            single_line (str): single line as read from input PVT file
            table_indices ([dict[str, list[int]]): dictionary to store the
                starting and ending line index of tables
            table_indices_dict (dict[str, dict[str, list[int]]]): dictionary to store the
                starting and ending line index of tables, for undersaturated tables
            l_index (int): current line index
            table_flag (dict[str, bool]): flag to tell if a table is currently being read (True) or not (False)
            table_has_endkeyword (bool): True if table name, e.g., PROPS, has end keyword, i.e., ENDPROPS, else False
            reading_flag (bool): True if any table is being read, otherwise False
            unsat_obj (dict[str, list[str]]): track saturation pressures from which undersaturated branches emanate

        Returns:
            bool: True if still reading table, but if identified the ending line index, return False
        """
        end_flag_found = False
        if table_key not in PVT_UNSAT_TABLE_INDICES:  # All tables except undersaturated tables
            table_name = table_key
            full_table_name = table_key
        else:  # Handle undersaturated tables
            if table_key == 'PRES':
                table_name = 'UNSATGAS'
                full_table_name = table_name + '_PRES'
            else:
                table_name = 'UNSATOIL'
                full_table_name = table_name + '_' + table_key
        # if not table_has_endkeyword and [i for i in single_line.split() if i in PVT_KEYWORDS]:
        if not table_has_endkeyword:
            for keyword in PVT_KEYWORDS:
                if nfo.check_token(keyword, single_line):
                    end_flag_found = True
        if table_has_endkeyword and nfo.check_token('END' + table_name, single_line):
            end_flag_found = True
        if (full_table_name in table_indices.keys() or full_table_name in table_indices_dict.keys()) and \
                end_flag_found and table_flag[table_name]:
            if table_key not in PVT_UNSAT_TABLE_INDICES:  # All tables except undersaturated tables
                table_indices[table_name][1] = l_index
            else:  # Handle undersaturated tables
                table_indices_dict[full_table_name][unsat_obj[table_key][-1]][1] = l_index
            table_flag[table_name] = False
            reading_flag = False
        return reading_flag

    def read_properties(self) -> None:
        """Read Nexus PVT file contents and populate the NexusPVTMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers used to record properties as we iterate through pvt file contents
        # Dictionary to record start and ending indices for tables
        pvt_table_indices: dict[str, list[int]] = {}
        pvt_table_indices_dict: dict[str, dict[str, list[int]]] = {}
        # Flag to tell when to start reading a table
        start_reading_table: bool = False
        trans_flag = False
        # Dictionary of flags indicating which tables are being read
        table_being_read: dict[str, bool] = {}
        for table_name in PVT_ALL_TABLE_KEYWORDS:
            table_being_read[table_name] = False
        # Dictionary of lists to track saturation pressures from which undersaturated branches emanate
        pvt_unsat_keys: dict[str, list[str]] = {}
        for indx in PVT_UNSAT_TABLE_INDICES:
            pvt_unsat_keys[indx] = []

        line_indx = 0
        for line in file_as_list:

            # Determine PVT type, i.e., BLACKOIL, WATEROIL, EOS, etc.
            for pvt_type in PVT_TYPE_KEYWORDS:
                if nfo.check_token(pvt_type, line):
                    self.pvt_type = pvt_type

            # Extract blackoil fluid density parameters
            for fluid_param in PVT_BLACKOIL_PRIMARY_KEYWORDS:
                if nfo.check_token(fluid_param, line):
                    self.properties[fluid_param] = float(nfo.get_expected_token_value(
                        fluid_param, line, file_as_list, custom_message=f"Property {fluid_param} does \
                        not have a numerical value."))
            if nfo.check_token('DRYGAS_MFP', line):
                self.properties['DRYGAS_MFP'] = True

            # For EOS or compositional models, get required parameters
            if nfo.check_token('NHC', line):  # Get number of components
                self.eos_nhc = int(nfo.get_expected_token_value('NHC', line,
                                                                file_as_list,
                                                                custom_message="Property NHC does not \
                                                                have a numerical value."))
            if nfo.check_token('COMPONENTS', line):  # Get NHC components
                elems = line.split()
                components_index = elems.index('COMPONENTS')
                if self.eos_nhc and self.eos_nhc > 0:
                    self.eos_components = elems[components_index + 1:components_index + 1 + int(self.eos_nhc)]
            if nfo.check_token('TEMP', line):  # Get default EOS temperature
                self.eos_temp = float(nfo.get_expected_token_value(
                    'TEMP', line, file_as_list, custom_message="Property TEMP does not have a numerical value."))

            # Check for EOS options
            if nfo.check_token('EOSOPTIONS', line):
                if nfo.get_expected_token_value('EOSOPTIONS', line, file_as_list) in PVT_EOS_METHODS:
                    self.eos_options['EOS_METHOD'] = nfo.get_expected_token_value('EOSOPTIONS', line, file_as_list)
                else:
                    self.eos_options['EOS_METHOD'] = 'PR'
            # Find EOS single-word options, like CAPILLARYFLASH and add to list
            if [i for i in line.split() if i in PVT_EOSOPTIONS_PRIMARY_WORDS]:
                if 'EOS_OPT_PRIMARY_LIST' not in self.eos_options.keys():
                    self.eos_options['EOS_OPT_PRIMARY_LIST'] = []
                if not isinstance(self.eos_options['EOS_OPT_PRIMARY_LIST'], list):
                    raise ValueError(f"EOS_OPT_PRIMARY_LIST should be a list, instead \
                                     got {self.eos_options['EOS_OPT_PRIMARY_LIST']}")
                self.eos_options['EOS_OPT_PRIMARY_LIST'].extend([i for i in line.split() if i
                                                                 in PVT_EOSOPTIONS_PRIMARY_WORDS])
            # Find EOS key-value pairs, like LI_FACT 0.9 or FUGERR 5
            if [i for i in line.split() if i in PVT_EOSOPTIONS_PRIMARY_KEYS_FLOAT]:
                for key in PVT_EOSOPTIONS_PRIMARY_KEYS_FLOAT:
                    if nfo.check_token(key, line):
                        self.eos_options[key] = float(nfo.get_expected_token_value(key, line, file_as_list))
            if [i for i in line.split() if i in PVT_EOSOPTIONS_PRIMARY_KEYS_INT]:
                for key in PVT_EOSOPTIONS_PRIMARY_KEYS_INT:
                    if nfo.check_token(key, line):
                        self.eos_options[key] = int(nfo.get_expected_token_value(key, line, file_as_list))
            # Read TRANSITION, TRANS_TEST and PHASEID eos options, if present
            primary_keys2populate = ['TRANSITION', 'TRANS_TEST', 'PHASEID']
            primary_keys2populate_defaults = ['TEST', 'INCRP', '']
            secondary_keys = [PVT_EOSOPTIONS_TRANS_KEYS, PVT_EOSOPTIONS_TRANS_TEST_KEYS, PVT_EOSOPTIONS_PHASEID_KEYS]
            if [i for i in line.split() if i in primary_keys2populate]:
                trans_flag = True
            if trans_flag:
                for index in range(len(primary_keys2populate)):
                    pkey = primary_keys2populate[index]
                    p2key = primary_keys2populate_defaults[index]
                    sec_key = secondary_keys[index]
                    self.__populate_eos_opts_to_tertiary_keys(pkey, p2key, line, file_as_list, sec_key)
            # Read TRANS_OPTIMIZATION eos options, if present
            if nfo.check_token('TRANS_OPTIMIZATION', line):
                new_dict: dict[str, float] = {}
                for tert_key in PVT_EOSOPTIONS_TERTIARY_KEYS:
                    if nfo.check_token(tert_key, line):
                        potential_value = float(nfo.get_expected_token_value(tert_key, line, file_as_list))
                        if isinstance(potential_value, float):
                            new_dict[tert_key] = potential_value
                            self.eos_options['TRANS_OPTIMIZATION'] = new_dict

            # Identify beginning and ending line indices for different kinds tables in PVT file
            if start_reading_table:  # Figure out ending line indices for tables
                for table in PVT_TABLES_WITHOUT_ENDWORDS:
                    start_reading_table = self.__find_pvt_table_ending_index(table, line, pvt_table_indices,
                                                                             pvt_table_indices_dict,
                                                                             line_indx, table_being_read, False,
                                                                             start_reading_table)
                for table in PVT_TABLES_WITH_ENDWORDS:
                    start_reading_table = self.__find_pvt_table_ending_index(table, line, pvt_table_indices,
                                                                             pvt_table_indices_dict,
                                                                             line_indx, table_being_read, True,
                                                                             start_reading_table)
                for table in PVT_UNSAT_TABLE_INDICES:
                    start_reading_table = self.__find_pvt_table_ending_index(table, line, pvt_table_indices,
                                                                             pvt_table_indices_dict,
                                                                             line_indx, table_being_read, False,
                                                                             start_reading_table, pvt_unsat_keys)
            # Figure out beginning line indices for tables
            table_found = False
            for table in PVT_TABLE_KEYWORDS + PVT_UNSAT_TABLE_INDICES:
                if table in PVT_UNSAT_TABLE_INDICES:  # Work on undersaturated tables
                    new_line_indx = self.__find_pvt_table_starting_index(table, line, file_as_list, pvt_table_indices,
                                                                         pvt_table_indices_dict, table_being_read,
                                                                         line_indx, pvt_unsat_keys)
                else:  # All other tables
                    new_line_indx = self.__find_pvt_table_starting_index(table, line, file_as_list, pvt_table_indices,
                                                                         pvt_table_indices_dict, table_being_read,
                                                                         line_indx)
                if new_line_indx is not None:
                    line_indx = new_line_indx
                    table_found = True
                    break
            if table_found:
                continue
            # Check if this line represents the header of a PVT table
            table_header_row_options = ['PRES', 'DP', 'RV', 'INDEX', 'COMPONENT']
            header_row_flag = False
            for hr in table_header_row_options:
                if nfo.check_token(hr, line):
                    header_row_flag = True
            reading_a_table_flag = False
            for table_name in PVT_ALL_TABLE_KEYWORDS:
                if table_being_read[table_name]:
                    reading_a_table_flag = True
            if header_row_flag and reading_a_table_flag and not start_reading_table:
                start_reading_table = True  # If header row, start reading

            line_indx += 1

        # Read in tables
        for key in pvt_table_indices.keys():
            if key == 'IGS_CP':
                self.properties[key] = nfo.read_table_to_df(file_as_list[
                    pvt_table_indices[key][0]:pvt_table_indices[key][1]], noheader=True)
            else:
                self.properties[key] = nfo.read_table_to_df(file_as_list[
                    pvt_table_indices[key][0]:pvt_table_indices[key][1]])
        for key in pvt_table_indices_dict.keys():
            self.properties[key] = {}
            for subkey in pvt_table_indices_dict[key].keys():
                property_key = self.properties[key]
                if not isinstance(property_key, dict):
                    raise ValueError(f"Property is not a dictionary: {str(self.properties[key])}")
                property_key[subkey] = nfo.read_table_to_df(file_as_list[
                                                            pvt_table_indices_dict[key][subkey][0]:
                                                            pvt_table_indices_dict[key][subkey][1]])



"""A class representing a single completion from a Nexus wells file."""
from __future__ import annotations
from dataclasses import dataclass

from typing import Optional, Union

# Use correct Self type depending upon Python version
import sys

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.NexusEnums import DateFormatEnum
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ResSimpy.Completion import Completion
from ResSimpy.Nexus.DataModels.NexusRelPermEndPoint import NexusRelPermEndPoint
from ResSimpy.Utils.to_dict_generic import to_dict


@dataclass(kw_only=True, repr=False)
class NexusCompletion(Completion):
    """A class representing a completion specific to a Nexus Model. Inherits from Completion.

    Additional Attributes:
        measured_depth (Optional[float]): Measured depth of a completion. 'MD' in Nexus
        well_indices (Optional[float]): Well index used to calculate performance of the completion. 'WI' in Nexus
        partial_perf (Optional[float]): Partial penetration factor. 'PPERF' in Nexus
        cell_number (Optional[int]): cell number for the completion in unstructured grids. 'CELL' in Nexus
        portype (Optional[str]): indicates the pore type for the completion FRACTURE OR MATRIX. 'PORTYPE' in Nexus
        sector (None | str | int): the section of the wellbore to which this completion flows. 'SECT' in Nexus
        khmult (Optional[float]): the multiplier that is applied to the permeability-thickness. 'KHMULT' in Nexus.
    """

    __measured_depth: Optional[float] = None
    __well_indices: Optional[float] = None
    __partial_perf: Optional[float] = None
    __cell_number: Optional[int] = None
    __bore_radius: Optional[float] = None
    __portype: Optional[str] = None
    __fracture_mult: Optional[float] = None
    __sector: Union[None, str, int] = None
    __well_group: Optional[str] = None
    __zone: Optional[int] = None
    __angle_open_flow: Optional[float] = None
    __temperature: Optional[float] = None
    __flowsector: Optional[int] = None
    __parent_node: Optional[str] = None
    __mdcon: Optional[float] = None
    __pressure_avg_pattern: Optional[int] = None
    __length: Optional[float] = None
    __permeability: Optional[float] = None
    __non_darcy_model: Optional[str] = None
    __comp_dz: Optional[float] = None
    __layer_assignment: Optional[int] = None
    __polymer_block_radius: Optional[float] = None
    __polymer_well_radius: Optional[float] = None
    __rel_perm_end_point: Optional[NexusRelPermEndPoint] = None
    __kh_mult: Optional[float] = None

    def __init__(self, date: str, i: Optional[int] = None, j: Optional[int] = None, k: Optional[int] = None,
                 skin: Optional[float] = None, depth: Optional[float] = None, well_radius: Optional[float] = None,
                 x: Optional[float] = None, y: Optional[float] = None, angle_a: Optional[float] = None,
                 angle_v: Optional[float] = None, grid: Optional[str] = None, measured_depth: Optional[float] = None,
                 well_indices: Optional[float] = None, depth_to_top: Optional[float] = None,
                 depth_to_bottom: Optional[float] = None, depth_to_top_str: Optional[str] = None,
                 depth_to_bottom_str: Optional[str] = None, rel_perm_method: Optional[int] = None,
                 dfactor: Optional[float] = None, status: Optional[str] = None, partial_perf: Optional[float] = None,
                 cell_number: Optional[int] = None, perm_thickness_ovr: Optional[float] = None,
                 peaceman_well_block_radius: Optional[float] = None, fracture_mult: Optional[float] = None,
                 sector: Union[None, str, int] = None, well_group: Optional[str] = None, zone: Optional[int] = None,
                 angle_open_flow: Optional[float] = None, temperature: Optional[float] = None,
                 flowsector: Optional[int] = None, parent_node: Optional[str] = None, mdcon: Optional[float] = None,
                 pressure_avg_pattern: Optional[int] = None, length: Optional[float] = None,
                 permeability: Optional[float] = None, non_darcy_model: Optional[str] = None,
                 comp_dz: Optional[float] = None, layer_assignment: Optional[int] = None,
                 polymer_bore_radius: Optional[float] = None, polymer_well_radius: Optional[float] = None,
                 portype: Optional[str] = None, rel_perm_end_point: Optional[NexusRelPermEndPoint] = None,
                 kh_mult: Optional[float] = None,
                 date_format: Optional[DateFormatEnum.DateFormat] = None,
                 start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None,
                 ) -> None:

        self.__measured_depth = measured_depth

        self.__well_indices = well_indices  # TODO: rename this to singular
        self.__partial_perf = partial_perf
        self.__depth_to_top_str = depth_to_top_str
        self.__depth_to_bottom_str = depth_to_bottom_str
        self.__cell_number = cell_number
        self.__fracture_mult = fracture_mult
        self.__sector = sector
        self.__well_group = well_group
        self.__zone = zone
        self.__angle_open_flow = angle_open_flow
        self.__temperature = temperature
        self.__flowsector = flowsector
        self.__parent_node = parent_node
        self.__mdcon = mdcon
        self.__pressure_avg_pattern = pressure_avg_pattern
        self.__length = length
        self.__permeability = permeability
        self.__non_darcy_model = non_darcy_model
        self.__comp_dz = comp_dz
        self.__layer_assignment = layer_assignment
        self.__polymer_block_radius = polymer_bore_radius
        self.__polymer_well_radius = polymer_well_radius
        self.__portype = portype
        self.__rel_perm_end_point = rel_perm_end_point
        self.__kh_mult = kh_mult

        super().__init__(date=date, i=i, j=j, k=k, skin=skin, depth=depth, well_radius=well_radius, x=x, y=y,
                         angle_a=angle_a, angle_v=angle_v, grid=grid, depth_to_top=depth_to_top,
                         depth_to_bottom=depth_to_bottom, perm_thickness_ovr=perm_thickness_ovr, dfactor=dfactor,
                         rel_perm_method=rel_perm_method, status=status, date_format=date_format, start_date=start_date,
                         unit_system=unit_system, peaceman_well_block_radius=peaceman_well_block_radius)

    @property
    def measured_depth(self):
        return self.__measured_depth

    @property
    def well_indices(self):
        return self.__well_indices

    @property
    def partial_perf(self):
        return self.__partial_perf

    @property
    def cell_number(self):
        return self.__cell_number

    @property
    def portype(self):
        return self.__portype

    @property
    def fracture_mult(self):
        return self.__fracture_mult

    @property
    def sector(self):
        return self.__sector

    @property
    def well_group(self):
        return self.__well_group

    @property
    def zone(self):
        return self.__zone

    @property
    def angle_open_flow(self):
        return self.__angle_open_flow

    @property
    def temperature(self):
        return self.__temperature

    @property
    def flowsector(self):
        return self.__flowsector

    @property
    def parent_node(self):
        return self.__parent_node

    @property
    def mdcon(self):
        return self.__mdcon

    @property
    def pressure_avg_pattern(self):
        return self.__pressure_avg_pattern

    @property
    def length(self):
        return self.__length

    @property
    def permeability(self):
        return self.__permeability

    @property
    def non_darcy_model(self):
        return self.__non_darcy_model

    @property
    def comp_dz(self):
        return self.__comp_dz

    @property
    def layer_assignment(self):
        return self.__layer_assignment

    @property
    def polymer_block_radius(self):
        return self.__polymer_block_radius

    @property
    def polymer_well_radius(self):
        return self.__polymer_well_radius

    @property
    def rel_perm_end_point(self):
        return self.__rel_perm_end_point

    @property
    def kh_mult(self):
        return self.__kh_mult

    @property
    def depth_to_top_str(self):
        return self.__depth_to_top_str

    @property
    def depth_to_bottom_str(self):
        return self.__depth_to_bottom_str

    def to_dict(self, keys_in_keyword_style: bool = False, add_date=True, add_units=False, include_nones=True) -> \
            dict[str, None | str | int | float]:

        # overwrite add_units to be False for completions as they currently do not contain units.
        attribute_dict = to_dict(self, keys_in_keyword_style, add_date, add_units=add_units,
                                 include_nones=include_nones)

        attribute_dict.update(super().to_dict(add_units=False))
        if self.rel_perm_end_point is not None:
            attribute_dict.update(self.rel_perm_end_point.to_dict())
        return attribute_dict

    @staticmethod
    def completion_is_perforation(completion: NexusCompletion) -> bool:
        """Determines if the supplied completion is a perforation or not."""

        # If we don't have any non-none values for these properties, assume the default values, which mean that the
        # layer is perforated
        if completion.partial_perf is None and completion.well_indices is None and completion.status is None \
                and completion.kh_mult is None and completion.perm_thickness_ovr is None:
            return True

        return ((completion.partial_perf is None or completion.partial_perf > 0) and
                (completion.well_indices is None or completion.well_indices > 0) and
                (completion.perm_thickness_ovr is None or completion.perm_thickness_ovr > 0) and
                (completion.kh_mult is None or completion.kh_mult > 0) and
                (completion.length is None or completion.length > 0) and
                (completion.status != 'OFF'))

    @staticmethod
    def completion_is_shutin(completion: NexusCompletion) -> bool:
        """Determines if the supplied completion is a shut-in or not."""
        return not NexusCompletion.completion_is_perforation(completion)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Returns a dictionary of mapping from nexus keyword to attribute name."""

        nexus_mapping: dict[str, tuple[str, type]] = {
            'IW': ('i', int),
            'JW': ('j', int),
            'L': ('k', int),
            'MD': ('measured_depth', float),
            'SKIN': ('skin', float),
            'DEPTH': ('depth', float),
            'X': ('x', float),
            'Y': ('y', float),
            'ANGLA': ('angle_a', float),
            'ANGLV': ('angle_v', float),
            'GRID': ('grid', str),
            'WI': ('well_indices', float),
            'DTOP': ('depth_to_top', float),
            'DBOT': ('depth_to_bottom', float),
            'RADW': ('well_radius', float),
            'PPERF': ('partial_perf', float),
            'CELL': ('cell_number', int),
            'KH': ('perm_thickness_ovr', float),
            'D': ('dfactor', float),
            'IRELPM': ('rel_perm_method', int),
            'STAT': ('status', str),
            'RADB': ('peaceman_well_block_radius', float),
            'PORTYPE': ('portype', str),
            'FM': ('fracture_mult', float),
            'SECT': ('sector', int),
            'GROUP': ('well_group', str),
            'ZONE': ('zone', int),
            'ANGLE': ('angle_open_flow', float),
            'TEMP': ('temperature', float),
            'FLOWSECT': ('flowsector', int),
            'PARENT': ('parent_node', str),
            'MDCON': ('mdcon', float),
            'IPTN': ('pressure_avg_pattern', int),
            'LENGTH': ('length', float),
            'K': ('permeability', float),
            'ND': ('non_darcy_model', str),
            'DZ': ('comp_dz', float),
            'LAYER': ('layer_assignment', int),
            'RADBP': ('polymer_block_radius', float),
            'RADWP': ('polymer_well_radius', float),
            'KHMULT': ('kh_mult', float),
        }

        return nexus_mapping

    @staticmethod
    def valid_attributes() -> list[str]:
        """Lists all possible attributes for the object that relate to a Nexus keyword."""
        return [v[0] for v in NexusCompletion.get_keyword_mapping().values()]

    @classmethod
    def from_dict(cls, input_dictionary: dict[str, None | float | int | str], date_format: DateFormat) -> Self:
        """Generates a NexusCompletion from a dictionary."""
        for input_attr in input_dictionary:
            if input_attr == 'date' or input_attr == 'unit_system' or input_attr == 'date_format':
                continue
            elif input_attr not in cls.valid_attributes():
                raise AttributeError(f'Unexpected keyword "{input_attr}" found within {input_dictionary}')
        date = input_dictionary.get('date', None)
        date_format_str = input_dictionary.get('date_format')
        if date_format_str is not None and isinstance(date_format_str, str):
            converted_date_format_str = date_format_str.replace('/', '_')
            completion_date_format = DateFormat[converted_date_format_str]
        else:
            completion_date_format = date_format
        if date is None:
            raise AttributeError(f'No date provided for the completion, instead got {date=}')

        date = str(date)
        constructed_class = cls(date=date, date_format=completion_date_format)
        constructed_class.update(input_dictionary)
        return constructed_class

    def update(self, input_dictionary: dict[str, None | float | int | str]) -> None:
        """Updates a completion based on a dictionary of attributes."""
        for k, v in input_dictionary.items():
            if v is None:
                continue
            if hasattr(self, '_NexusCompletion__' + k):
                setattr(self, '_NexusCompletion__' + k, v)
            elif hasattr(super(), '_Completion__' + k):
                setattr(self, '_Completion__' + k, v)



"""Class to hold Nexus Valve properties."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.valve_keywords import VALVE_TABLE_KEYWORDS, VALVE_RATE_KEYWORDS
from ResSimpy.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import HydraulicsUnits
from ResSimpy.Enums.UnitsEnum import UnitSystem, TemperatureUnits
from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusValveMethod(DynamicProperty):
    """Class to hold Nexus Valve properties.

    Attributes:
        file (NexusFile): Nexus valve properties file object
        input_number (int): Valve properties method number in Nexus fcs file
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific valve properties method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile
    properties: dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                     dict[str, Union[float, pd.DataFrame]]]] = field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'VC': ('valve_coefficient', float),
        }
        return keywords

    @property
    def units(self) -> HydraulicsUnits:
        """Returns the attribute to unit map for the Valve method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return HydraulicsUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with valve data in Nexus file format."""
        printable_str = ''
        valve_dict = self.properties
        for key, value in valve_dict.items():
            if key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif isinstance(value, pd.DataFrame):
                table_text = f'{key}'
                if key in ['VALVE', 'ICV']:
                    for rate_key in VALVE_RATE_KEYWORDS:
                        if 'DP_RATE' in valve_dict.keys() and rate_key == valve_dict['DP_RATE']:
                            table_text += f' {rate_key}'
                printable_str += f'{table_text}\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n'
                printable_str += 'END'+key+'\n'
                printable_str += '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
            elif key not in ['DP_RATE']:
                printable_str += f'{key} {value}\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus valve file contents and populate the NexusValveMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers to use to record properties as we iterate through valve file contents
        # Dictionary to record start and ending indices for tables
        valve_table_indices: dict[str, list[int]] = {}
        # Dictionary of flags indicating which tables are being read
        table_being_read: dict[str, bool] = {}
        for table_name in VALVE_TABLE_KEYWORDS:
            table_being_read[table_name] = False

        line_indx = 0
        for line in file_as_list:

            # Find beginning and ending indices of tables
            for table_key in VALVE_TABLE_KEYWORDS:
                if nfo.check_token(table_key, line):
                    table_being_read[table_key] = True
                    valve_table_indices[table_key] = [line_indx + 1, len(file_as_list)]
                    # Find rate keyword if present
                    for rate_key in VALVE_RATE_KEYWORDS:
                        if nfo.check_token(rate_key, line):
                            self.properties['DP_RATE'] = rate_key  # Rate used to determine pressure drop across valve
                if table_being_read[table_key] and nfo.check_token('END' + table_key, line):
                    table_being_read[table_key] = False
                    valve_table_indices[table_key][1] = line_indx

            line_indx += 1

        # Read in tables
        for key in valve_table_indices.keys():
            self.properties[key] = nfo.read_table_to_df(file_as_list[
                                                        valve_table_indices[key][0]:valve_table_indices[key][1]])




ROCK_KEYWORDS_VALUE_FLOAT = ['CR', 'KP', 'PREF', 'TOLREV_P', 'TOLREV_SW', 'CR_REPRESSURE', 'KP_REPRESSURE',
                             'DPMIN', 'CREEPB', 'CREEPC', 'CREEPM']
ROCK_KEYWORDS_VALUE_STR = ['REVERSIBLE', 'IRREVERSIBLE']
ROCK_REV_IRREV_OPTIONS = ['ALL', 'CMTONLY', 'WIRCTONLY']
ROCK_SINGLE_KEYWORDS = ['COMPR', 'KPMULT', 'CREEP', 'TAMULTONLY']
ROCK_CMT = ['CMT']
ROCK_WATER_INDUCED_CMT = ['WIRCT', 'SWINIT']
ROCK_ALL_TABLE_KEYWORDS = ROCK_CMT + ROCK_WATER_INDUCED_CMT

# Apparent Permeability in Nano-porosity Rocks (Knudsen diffusion/slippage) options
ROCK_KNUDSEN_DIFF_KEYWORDS = ['JAVADPOUR', 'CIVAN', 'KNUDSEN']
ROCK_KNUDSEN_DIFF_OPTIONS = ['R', 'ALPHA', 'BETA', 'A', 'B']

ROCK_KEYWORDS = ROCK_KEYWORDS_VALUE_FLOAT + ROCK_KEYWORDS_VALUE_STR + ROCK_SINGLE_KEYWORDS + \
    ROCK_ALL_TABLE_KEYWORDS + ROCK_REV_IRREV_OPTIONS + ROCK_KNUDSEN_DIFF_KEYWORDS + ROCK_KNUDSEN_DIFF_OPTIONS



from typing import Any

import numpy as np
import pandas as pd


def obj_to_dataframe(list_objs: list[Any]) -> pd.DataFrame:
    """Returns a dataframe representing the attributes of the object as rows of the dataframe
    Requires a "to_dict" method for each object.
    """
    df_store = pd.DataFrame([x.to_dict() for x in list_objs])
    df_store = df_store.fillna(value=np.nan)
    df_store = df_store.dropna(axis=1, how='all')
    return df_store



"""File object for handling FCS top level files for Nexus."""
from __future__ import annotations
from dataclasses import dataclass, field
import os
import warnings

from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from typing import Optional

# Use correct Self type depending upon Python version
import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ResSimpy.Utils.factory_methods import get_empty_dict_int_nexus_file, get_empty_list_str, \
    get_empty_list_nexus_file
from ResSimpy.Nexus.NexusKeywords.fcs_keywords import FCS_KEYWORDS
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Utils.generic_repr import generic_repr, generic_str
from datetime import datetime


@dataclass(kw_only=True)
class FcsNexusFile(NexusFile):
    restart_file: Optional[NexusFile] = None
    structured_grid_file: Optional[NexusFile] = None
    options_file: Optional[NexusFile] = None
    runcontrol_file: Optional[NexusFile] = None
    override_file: Optional[NexusFile] = None
    eos_default_file: Optional[NexusFile] = None
    well_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    surface_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    rock_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    relperm_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    pvt_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    water_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    equil_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    tracer_init_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    aquifer_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    hyd_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    valve_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    separator_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    ipr_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    gas_lift_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    pump_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    compressor_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    choke_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    icd_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    esp_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    polymer_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    adsorption_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    flux_in_files: Optional[dict[int, NexusFile]] = field(default_factory=get_empty_dict_int_nexus_file)
    files_info: list[tuple[Optional[str], Optional[str], Optional[datetime]]]

    def __init__(
            self, location: str,
            include_locations: Optional[list[str]] = None,
            origin: Optional[str] = None,
            include_objects: Optional[list[NexusFile]] = None,
            file_content_as_list: Optional[list[str]] = None,
            restart_file: Optional[NexusFile] = None,
            structured_grid_file: Optional[NexusFile] = None,
            options_file: Optional[NexusFile] = None,
            runcontrol_file: Optional[NexusFile] = None,
            override_file: Optional[NexusFile] = None,
            eos_default_file: Optional[NexusFile] = None,
            well_files: Optional[dict[int, NexusFile]] = None,
            surface_files: Optional[dict[int, NexusFile]] = None,
            rock_files: Optional[dict[int, NexusFile]] = None,
            relperm_files: Optional[dict[int, NexusFile]] = None,
            pvt_files: Optional[dict[int, NexusFile]] = None,
            water_files: Optional[dict[int, NexusFile]] = None,
            equil_files: Optional[dict[int, NexusFile]] = None,
            tracer_init_files: Optional[dict[int, NexusFile]] = None,
            aquifer_files: Optional[dict[int, NexusFile]] = None,
            hyd_files: Optional[dict[int, NexusFile]] = None,
            valve_files: Optional[dict[int, NexusFile]] = None,
            separator_files: Optional[dict[int, NexusFile]] = None,
            ipr_files: Optional[dict[int, NexusFile]] = None,
            gas_lift_files: Optional[dict[int, NexusFile]] = None,
            pump_files: Optional[dict[int, NexusFile]] = None,
            compressor_files: Optional[dict[int, NexusFile]] = None,
            choke_files: Optional[dict[int, NexusFile]] = None,
            icd_files: Optional[dict[int, NexusFile]] = None,
            esp_files: Optional[dict[int, NexusFile]] = None,
            polymer_files: Optional[dict[int, NexusFile]] = None,
            adsorption_files: Optional[dict[int, NexusFile]] = None,
            flux_in_files: Optional[dict[int, NexusFile]] = None
    ) -> None:
        self.restart_file = restart_file
        self.structured_grid_file = structured_grid_file
        self.options_file = options_file
        self.runcontrol_file = runcontrol_file
        self.override_file = override_file
        self.eos_default_file = eos_default_file
        self.surface_files = surface_files
        self.well_files = well_files if well_files is not None else get_empty_dict_int_nexus_file()
        self.rock_files = rock_files if rock_files is not None else get_empty_dict_int_nexus_file()
        self.relperm_files = relperm_files if relperm_files is not None else get_empty_dict_int_nexus_file()
        self.pvt_files = pvt_files if pvt_files is not None else get_empty_dict_int_nexus_file()
        self.water_files = water_files if water_files is not None else get_empty_dict_int_nexus_file()
        self.equil_files = equil_files if equil_files is not None else get_empty_dict_int_nexus_file()
        self.tracer_init_files = tracer_init_files if tracer_init_files is not None else get_empty_dict_int_nexus_file()
        self.aquifer_files = aquifer_files if aquifer_files is not None else get_empty_dict_int_nexus_file()
        self.hyd_files = hyd_files if hyd_files is not None else get_empty_dict_int_nexus_file()
        self.valve_files = valve_files if valve_files is not None else get_empty_dict_int_nexus_file()
        self.separator_files = separator_files if separator_files is not None else get_empty_dict_int_nexus_file()
        self.ipr_files = ipr_files if ipr_files is not None else get_empty_dict_int_nexus_file()
        self.gas_lift_files = gas_lift_files if gas_lift_files is not None else get_empty_dict_int_nexus_file()
        self.pump_files = pump_files if pump_files is not None else get_empty_dict_int_nexus_file()
        self.compressor_files = compressor_files if compressor_files is not None else get_empty_dict_int_nexus_file()
        self.choke_files = choke_files if choke_files is not None else get_empty_dict_int_nexus_file()
        self.icd_files = icd_files if icd_files is not None else get_empty_dict_int_nexus_file()
        self.esp_files = esp_files if esp_files is not None else get_empty_dict_int_nexus_file()
        self.polymer_files = polymer_files if polymer_files is not None else get_empty_dict_int_nexus_file()
        self.adsorption_files = adsorption_files if adsorption_files is not None else get_empty_dict_int_nexus_file()
        self.flux_in_files = flux_in_files if flux_in_files is not None else get_empty_dict_int_nexus_file()
        self.files_info = []
        super().__init__(location=location, include_locations=include_locations, origin=origin,
                         include_objects=include_objects, file_content_as_list=file_content_as_list)

    def __repr__(self) -> str:
        return generic_repr(self)

    def __str__(self) -> str:
        return generic_str(self)

    @classmethod
    def generate_fcs_structure(cls, fcs_file_path: str, recursive: bool = True) -> Self:
        """Creates an instance of the FcsNexusFile, populates it through looking through the different keywords \
            in the FCS and assigning the paths to objects.

        Args:
        ----
            fcs_file_path (str): path to the fcs file of interest
            recursive (bool, optional): Whether the NexusFile structure will be recursively created. Defaults to True.

        Raises:
        ------
            FileNotFoundError: if the fcs file cannot be found
            ValueError: if no content can be found within the fcsfile

        Returns:
        -------
            FcsNexusFile: instance of a FcsNexusFile for a given fcs file path
        """
        fcs_file = cls(location=fcs_file_path)
        fcs_file.include_objects = get_empty_list_nexus_file()
        fcs_file.file_content_as_list = get_empty_list_str()
        fcs_file.include_locations = get_empty_list_str()

        # guard against bad links/empty files:
        if not os.path.isfile(fcs_file_path):
            raise FileNotFoundError(f'fcs file not found for path {fcs_file_path}')
        origin_path = fcs_file_path
        fcs_nexus_file = NexusFile.generate_file_include_structure(
            fcs_file_path, origin=None)
        fcs_file.files_info.append((fcs_nexus_file.location, fcs_nexus_file.linked_user,
                                    fcs_nexus_file.last_modified))
        flat_fcs_file_content = fcs_nexus_file.get_flat_list_str_file
        if flat_fcs_file_content is None or fcs_file.file_content_as_list is None:
            raise ValueError(f'FCS file not found, no content for {fcs_file_path=}')
        fcs_file.file_content_as_list = flat_fcs_file_content
        for i, line in enumerate(flat_fcs_file_content):
            if not nfo.nexus_token_found(line, valid_list=FCS_KEYWORDS):
                continue
            key = nfo.get_next_value(start_line_index=i, file_as_list=flat_fcs_file_content, search_string=line)
            if key is None:
                warnings.warn(f'get next value failed to find a suitable token in {line}')
                continue
            key = key.upper()
            value = nfo.get_token_value(key, line, flat_fcs_file_content[i::])
            if value is None:
                warnings.warn(f'No value found for {key}, skipping file')
                continue
            # TODO handle methods / sets instead of getting full file path
            if key in cls.fcs_keyword_map_multi():
                # for keywords that have multiple methods we store the value in a dictionary
                # with the method number and the NexusFile object
                _, method_string, method_number, value = (
                    nfo.get_multiple_sequential_values(flat_fcs_file_content[i::], 4, ['NORPT'])
                )
                full_file_path = nfo.get_full_file_path(value, origin_path)
                nexus_file = NexusFile.generate_file_include_structure(
                    value, origin=fcs_file_path, recursive=recursive, top_level_file=True)
                fcs_property = getattr(fcs_file, cls.fcs_keyword_map_multi()[key])
                # manually initialise if the property is still a None after class instantiation
                if fcs_property is None:
                    fcs_property = get_empty_dict_int_nexus_file()
                # shallow copy to maintain list elements pointing to nexus_file that are
                # stored in the file_content_as_list
                fcs_property_list = fcs_property.copy()
                fcs_property_list.update({int(method_number): nexus_file})
                # set the attribute in the FcsNexusFile instance
                setattr(fcs_file, cls.fcs_keyword_map_multi()[key], fcs_property_list)
                fcs_file.include_objects.append(nexus_file)
                fcs_file.include_locations.append(full_file_path)
                fcs_file.files_info.append((nexus_file.location, nexus_file.linked_user,
                                            nexus_file.last_modified))
            elif key in cls.fcs_keyword_map_single():
                full_file_path = nfo.get_full_file_path(value, origin_path)
                nexus_file = NexusFile.generate_file_include_structure(
                    value, origin=fcs_file_path, recursive=recursive, top_level_file=True)
                setattr(fcs_file, cls.fcs_keyword_map_single()[key], nexus_file)
                fcs_file.include_objects.append(nexus_file)
                fcs_file.include_locations.append(full_file_path)
                fcs_file.files_info.append((nexus_file.location, nexus_file.linked_user,
                                            nexus_file.last_modified))
            else:
                continue
        return fcs_file

    @staticmethod
    def fcs_keyword_map_single() -> dict[str, str]:
        keyword_map = {
            'STRUCTURED_GRID': 'structured_grid_file',
            'OPTIONS': 'options_file',
            'RUNCONTROL': 'runcontrol_file',
            'OVERRIDE': 'override_file',
            'EOS_DEFAULTS': 'eos_default_file',
        }
        return keyword_map

    @staticmethod
    def fcs_keyword_map_multi() -> dict[str, str]:
        keyword_map = {
            'EQUIL': 'equil_files',
            'ROCK': 'rock_files',
            'RELPM': 'relperm_files',
            'SURFACE': 'surface_files',
            'WELLS': 'well_files',
            'PVT': 'pvt_files',
            'AQUIFER': 'aquifer_files',
            'HYD': 'hyd_files',
            'VALVE': 'valve_files',
            'WATER': 'water_files',
            'SEPARATOR': 'separator_files',
            'IPR': 'ipr_files',
            'GASLIFT': 'gas_lift_files',
            'PUMP': 'pump_files',
            'COMPRESSOR': 'compressor_files',
            'CHOKE': 'choke_files',
            'ICD': 'icd_files',
            'ESP': 'esp_files',
            'POLYMER': 'polymer_files',
            'ADSORPTION': 'adsorption_files',
            'FLUXIN': 'flux_in_files',
            'TRACER_INIT': 'tracer_init_files',
        }
        return keyword_map

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        single_keywords = {k: (v, type(NexusFile)) for k, v in FcsNexusFile.fcs_keyword_map_single().items()}
        multi_keywords = {k: (v, type(dict)) for k, v in FcsNexusFile.fcs_keyword_map_multi().items()}
        return_dict = dict(single_keywords, **multi_keywords)
        return return_dict

    def move_model_files(self, new_file_path: str, new_include_file_location: str) -> None:
        """Moves all the model files to a new location.

        Args:
            new_file_path (str): new file path for the fcs file e.g. /new_path/new_fcs_file.fcs
            new_include_file_location (str): new location for the included files either absolute or relative
            to the new fcs file path
        """
        # Take the original file, find which files have changed and write out those locations
        # figure out where to store the include files:
        file_directory = os.path.dirname(new_file_path)
        include_dir = new_include_file_location
        if not os.path.isabs(new_include_file_location):
            include_dir = os.path.join(file_directory, new_include_file_location)

        if not os.path.exists(include_dir):
            # make the folders if they don't already exist
            os.makedirs(include_dir)

        # Loop through all files in the model, writing out the contents
        for keyword, attr_name in self.fcs_keyword_map_single().items():
            file: None | NexusFile = getattr(self, attr_name, None)
            if file is None:
                # skip if there is no file
                continue
            include_name = os.path.join(include_dir, os.path.basename(file.location))
            file.write_to_file(include_name, write_includes=True, write_out_all_files=True)
            self.change_file_path(include_name, keyword)

        for keyword, attr_name in self.fcs_keyword_map_multi().items():
            file_dict: None | dict[int, NexusFile] = getattr(self, attr_name, None)
            if file_dict is None or len(file_dict) == 0:
                continue
            for method_number, file in file_dict.items():
                include_name = os.path.join(include_dir, os.path.basename(file.location))
                file.write_to_file(include_name, write_includes=True, write_out_all_files=True)
                self.change_file_path(include_name, keyword, method_number)

        # write out the final fcs file
        self.write_to_file(new_file_path, write_includes=False)

    def write_out_case(self, new_file_path: str, new_include_file_location: str, case_suffix: str) -> None:
        """Writes out a new simulator with only modified files. For use with creating multiple cases from a base case.

        Args:
            new_file_path (str): new file path for the fcs file e.g. /new_path/new_fcs_file.fcs
            new_include_file_location (str): new location for the included files either absolute or relative
            to the new fcs file path
            case_suffix (str): suffix to append to the end of the file name e.g. case_1
        """
        def new_include_file_name(file_name: str) -> str:
            """Returns the new include file name based on the original file name plus the suffix provided."""
            file_name = os.path.basename(file_name)
            file_name, file_extension = os.path.splitext(file_name)
            file_name = f'{file_name}_{case_suffix}{file_extension}'
            return os.path.join(new_include_file_location, file_name)

        file_directory = os.path.dirname(new_file_path)
        include_dir = new_include_file_location
        if not os.path.isabs(new_include_file_location):
            include_dir = os.path.join(file_directory, new_include_file_location)

        if not os.path.exists(include_dir):
            # make the folders if they don't already exist
            os.makedirs(include_dir)

        # Loop through all files in the model, writing out the contents
        for keyword, attr_name in self.fcs_keyword_map_single().items():
            file: None | NexusFile = getattr(self, attr_name, None)
            if file is None:
                # skip if there is no file
                continue
            include_write_name = file.location
            if file.file_modified:
                include_write_name = new_include_file_name(file.location)
                file.write_to_file(include_write_name, write_includes=True, write_out_all_files=False)
            self.change_file_path(include_write_name, keyword)

        for keyword, attr_name in self.fcs_keyword_map_multi().items():
            file_dict: None | dict[int, NexusFile] = getattr(self, attr_name, None)
            if file_dict is None or len(file_dict) == 0:
                continue
            for method_number, file in file_dict.items():
                include_write_name = file.location
                if file.file_modified:
                    include_write_name = new_include_file_name(file.location)
                    file.write_to_file(include_write_name, write_includes=True, write_out_all_files=False)
                self.change_file_path(include_write_name, keyword, method_number)
        self.write_to_file(new_file_path, write_includes=False)

    def update_model_files(self) -> None:
        """Updates all the modified files and the fcs in the model. Keeps file names and paths the same.
        Warning: this method overwrites the existing files!
        """
        # Loop through all files in the model, writing out the contents if they have been modified.
        for keyword, attr_name in self.fcs_keyword_map_single().items():
            file: None | NexusFile = getattr(self, attr_name, None)
            if file is None:
                # skip if there is no file
                continue
            if file.file_modified:
                file.write_to_file(write_includes=True, write_out_all_files=False, overwrite_file=True)

        for keyword, attr_name in self.fcs_keyword_map_multi().items():
            file_dict: None | dict[int, NexusFile] = getattr(self, attr_name, None)
            if file_dict is None or len(file_dict) == 0:
                continue
            for method_number, file in file_dict.items():
                if file.file_modified:
                    file.write_to_file(write_includes=True, write_out_all_files=False, overwrite_file=True)

        # write out the final fcs file
        self.write_to_file(self.location, write_includes=False, overwrite_file=True)

    def change_file_path(self, new_file_path: str, token: str, method_number: int | None = None) -> bool:
        """Switch the file path for a new file_path based on the value of the associated keyword in the fcs.

        Args:
            new_file_path (str): file path to replace the existing file path with.
            token (str): token indicating the
            method_number (int | None): method number for the file to replace.
            If None then will only look for keywords that follow directly after a token
        Returns:
            (bool): representing whether the file was successfully changed
        """
        file_changed = False
        if self.file_content_as_list is None:
            raise ValueError('No file content to change file path on.')
        for index, line in enumerate(self.get_flat_list_str_file):
            if not nfo.check_token(token, line):
                continue

            if method_number is not None:
                token_from_file, intermediate_word, method_num_in_file, path_to_replace \
                    = nfo.get_multiple_sequential_values([line], 4, ['NORPT'])
                if int(method_num_in_file) != method_number:
                    continue
            else:
                path_to_replace = nfo.get_expected_token_value(token, line, self.get_flat_list_str_file)
            # edit the file as list
            file_to_edit, index_to_mod = self.find_which_include_file(flattened_index=index)
            if file_to_edit.file_content_as_list is None:
                raise ValueError(f'No content found within {file_to_edit.location}')
            file_to_edit.file_content_as_list[index_to_mod] = line.replace(path_to_replace, new_file_path)
            file_changed = True
            self._file_modified_set(file_changed)
        return file_changed



"""A module containing the data objects for the Nexus Network."""
__all__ = ['NexusConstraint', 'NexusConstraints', 'NexusNode', 'NexusNodeConnection', 'NexusNodeConnections',
           'NexusNodes', 'NexusWellbore', 'NexusWellbores', 'NexusWellConnection', 'NexusWellConnections',
           'NexusWellhead', 'NexusWellheads']



FCS_KEYWORDS = ['RESERVOIR', 'EQUIL', 'EOS_DEFAULTS', 'TRACER_INIT', 'STRUCTURED_GRID', 'OPTIONS', 'ROCK', 'RELPM',
                'ADSORPTION', 'PVT', 'SEPARATOR', 'ALPHAF', 'POLYMER', 'WATER', 'AQUIFER', 'FLUXIN', 'RUNCONTROL',
                'WELLS', 'SURFACE', 'IPR', 'GASLIFT', 'VALVE', 'HYD', 'PUMP', 'COMPRESSOR', 'CHOKE', 'ICD', 'ESP']

# TODO: Not a complete list. Check manual for missing keywords.
# TODO: Delete the keywords that are not tokens (i.e. that are only values in a table)

FCS_ALL_KEYWORDS = ['ADSORPTION', 'ALPHA', 'ALPHAF', 'AQUIFER', 'AQUIFER_FILES', 'CASE', 'CHOKE', 'COMPRESSOR',
                    'DATEFORMAT', 'DD/MM/YYYY', 'DEFAULT_SUNITS', 'DEFAULT_UNITS', 'DESC', 'ENGLISH', 'EOS_DEFAULTS',
                    'EQUIL', 'ESP', 'FIELD', 'FLUXIN', 'GASLIFT', 'GRID_FILES', 'HYD', 'ICD',
                    'INITIALIZATION_FILES', 'IPR', 'LGR', 'METBAR', 'METHOD', 'MM/DD/YYYY', 'NET_METHOD_FILES', 'ON',
                    'OPTIONS', 'OVERRIDE', 'POLYMER', 'PPM', 'PROPERTY_FILES', 'PUMP', 'PVT', 'PVT_FILES',
                    'RECURRENT_FILES', 'RELPM', 'RESERVOIR', 'RESTART', 'ROCK', 'ROCK_FILES', 'RUNCONTROL',
                    'RUN_SUNITS',
                    'RUN_UNITS', 'SEPARATOR', 'STRUCTURED_GRID', 'SURFACE', 'TIME', 'TRACER_INIT', 'VALVE', 'VIP',
                    'WATER',
                    'WELLS', 'X']



"""Unit types for the attributes of a well completion."""
from typing import Mapping

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping
from ResSimpy.Units.Units import UnitDimension, Length, Temperature, Dimensionless, \
    Angle, PermeabilityThickness, Permeability, NonDarcySkin


class CompletionUnits(BaseUnitMapping):
    """Unit types for the attributes of a well completion."""
    attribute_map: Mapping[str, UnitDimension] = {
        'i': Dimensionless(),
        'j': Dimensionless(),
        'k': Dimensionless(),
        'x': Length(),
        'y': Length(),
        'depth': Length(),
        'depth_to_top': Length(),
        'depth_to_bottom': Length(),
        'well_radius': Length(),
        'skin': Dimensionless(),
        'angle_a': Angle(),
        'angle_v': Angle(),
        'grid': Dimensionless(),
        'perm_thickness_ovr': PermeabilityThickness(),
        'dfactor': NonDarcySkin(),
        'rel_perm_method': Dimensionless(),
        'status': Dimensionless(),
        'measured_depth': Length(),
        'well_indices': Dimensionless(),
        'partial_perf': Dimensionless(),
        'cell_number': Dimensionless(),
        'peaceman_well_block_radius': Length(),
        'portype': Dimensionless(),
        'facture_mult': Dimensionless(),
        'sector': Dimensionless(),
        'well_group': Dimensionless(),
        'zone': Dimensionless(),
        'angle_open_flow': Angle(),
        'temperature': Temperature(),
        'flowsector': Dimensionless(),
        'parent_node': Dimensionless(),
        'mdcon': Dimensionless(),
        'pressure_avg_pattern': Dimensionless(),
        'length': Length(),
        'permeability': Permeability(),
        'non_darcy_model': Dimensionless(),
        'comp_dz': Length(),
        'layer_assignment': Dimensionless(),
        'polymer_block_radius': Length(),
        'polymer_well_radius': Length(),
        'rel_perm_end_point': Dimensionless(),
        'kh_mult': Dimensionless(),
    }

    def __init__(self, unit_system: None | UnitSystem) -> None:
        super().__init__(unit_system=unit_system)

    @property
    def i(self) -> str:
        """Returns the unit for i."""
        return self.get_unit_for_attribute('i')

    @property
    def j(self) -> str:
        """Returns the unit for j."""
        return self.get_unit_for_attribute('j')

    @property
    def k(self) -> str:
        """Returns the unit for k."""
        return self.get_unit_for_attribute('k')

    @property
    def x(self) -> str:
        """Returns the unit for x."""
        return self.get_unit_for_attribute('x')

    @property
    def y(self) -> str:
        """Returns the unit for y."""
        return self.get_unit_for_attribute('y')

    @property
    def depth(self) -> str:
        """Returns the unit for depth."""
        return self.get_unit_for_attribute('depth')

    @property
    def depth_to_top(self) -> str:
        """Returns the unit for depth_to_top."""
        return self.get_unit_for_attribute('depth_to_top')

    @property
    def depth_to_bottom(self) -> str:
        """Returns the unit for depth_to_bottom."""
        return self.get_unit_for_attribute('depth_to_bottom')

    @property
    def well_radius(self) -> str:
        """Returns the unit for well_radius."""
        return self.get_unit_for_attribute('well_radius')

    @property
    def skin(self) -> str:
        """Returns the unit for skin."""
        return self.get_unit_for_attribute('skin')

    @property
    def angle_a(self) -> str:
        """Returns the unit for angle_a."""
        return self.get_unit_for_attribute('angle_a')

    @property
    def angle_v(self) -> str:
        """Returns the unit for angle_v."""
        return self.get_unit_for_attribute('angle_v')

    @property
    def grid(self) -> str:
        """Returns the unit for grid."""
        return self.get_unit_for_attribute('grid')

    @property
    def perm_thickness_ovr(self) -> str:
        """Returns the unit for perm_thickness_ovr."""
        return self.get_unit_for_attribute('perm_thickness_ovr')

    @property
    def dfactor(self) -> str:
        """Returns the unit for dfactor."""
        return self.get_unit_for_attribute('dfactor')

    @property
    def rel_perm_method(self) -> str:
        """Returns the unit for rel_perm_method."""
        return self.get_unit_for_attribute('rel_perm_method')

    @property
    def status(self) -> str:
        """Returns the unit for status."""
        return self.get_unit_for_attribute('status')

    @property
    def measured_depth(self) -> str:
        """Returns the unit for measured_depth."""
        return self.get_unit_for_attribute('measured_depth')

    @property
    def well_indices(self) -> str:
        """Returns the unit for well_indices."""
        return self.get_unit_for_attribute('well_indices')

    @property
    def partial_perf(self) -> str:
        """Returns the unit for partial_perf."""
        return self.get_unit_for_attribute('partial_perf')

    @property
    def cell_number(self) -> str:
        """Returns the unit for cell_number."""
        return self.get_unit_for_attribute('cell_number')

    @property
    def peaceman_well_block_radius(self) -> str:
        """Returns the unit for peaceman_well_block_radius."""
        return self.get_unit_for_attribute('peaceman_well_block_radius')

    @property
    def portype(self) -> str:
        """Returns the unit for portype."""
        return self.get_unit_for_attribute('portype')

    @property
    def facture_mult(self) -> str:
        """Returns the unit for facture_mult."""
        return self.get_unit_for_attribute('facture_mult')

    @property
    def sector(self) -> str:
        """Returns the unit for sector."""
        return self.get_unit_for_attribute('sector')

    @property
    def well_group(self) -> str:
        """Returns the unit for well_group."""
        return self.get_unit_for_attribute('well_group')

    @property
    def zone(self) -> str:
        """Returns the unit for zone."""
        return self.get_unit_for_attribute('zone')

    @property
    def angle_open_flow(self) -> str:
        """Returns the unit for angle_open_flow."""
        return self.get_unit_for_attribute('angle_open_flow')

    @property
    def temperature(self) -> str:
        """Returns the unit for temperature."""
        return self.get_unit_for_attribute('temperature')

    @property
    def flowsector(self) -> str:
        """Returns the unit for flowsector."""
        return self.get_unit_for_attribute('flowsector')

    @property
    def parent_node(self) -> str:
        """Returns the unit for parent_node."""
        return self.get_unit_for_attribute('parent_node')

    @property
    def mdcon(self) -> str:
        """Returns the unit for mdcon."""
        return self.get_unit_for_attribute('mdcon')

    @property
    def pressure_avg_pattern(self) -> str:
        """Returns the unit for pressure_avg_pattern."""
        return self.get_unit_for_attribute('pressure_avg_pattern')

    @property
    def length(self) -> str:
        """Returns the unit for length."""
        return self.get_unit_for_attribute('length')

    @property
    def permeability(self) -> str:
        """Returns the unit for permeability."""
        return self.get_unit_for_attribute('permeability')

    @property
    def non_darcy_model(self) -> str:
        """Returns the unit for non_darcy_model."""
        return self.get_unit_for_attribute('non_darcy_model')

    @property
    def comp_dz(self) -> str:
        """Returns the unit for comp_dz."""
        return self.get_unit_for_attribute('comp_dz')

    @property
    def layer_assignment(self) -> str:
        """Returns the unit for layer_assignment."""
        return self.get_unit_for_attribute('layer_assignment')

    @property
    def polymer_block_radius(self) -> str:
        """Returns the unit for polymer_block_radius."""
        return self.get_unit_for_attribute('polymer_block_radius')

    @property
    def polymer_well_radius(self) -> str:
        """Returns the unit for polymer_well_radius."""
        return self.get_unit_for_attribute('polymer_well_radius')

    @property
    def rel_perm_end_point(self) -> str:
        """Returns the unit for rel_perm_end_point."""
        return self.get_unit_for_attribute('rel_perm_end_point')

    @property
    def kh_mult(self) -> str:
        """Returns the unit for kh_mult."""
        return self.get_unit_for_attribute('kh_mult')



# TODO: Missing time data related values - RESTART, PLUS

WELLS_KEYWORDS = ['ADJPERF', 'ANGLA', 'ANGLE', 'ANGLV', 'CELL', 'CON', 'D', 'DATEFORMAT', 'DBOT', 'DEPTH', 'DKH',
                  'DKRG', 'DKRW', 'DSGR', 'DSWR', 'DTOP', 'DZ', 'ENDADJPERF', 'ENGLISH', 'FLOWSECT', 'FM', 'GRID',
                  'GROUP', 'IPTN', 'IRELPM', 'IW', 'JW', 'K', 'KH', 'KHMIN', 'KHMULT', 'KRG_SGRO', 'KRG_SGRW',
                  'KRG_SGU', 'KRO_SGL', 'KRO_SGR', 'KRO_SWL', 'KRO_SWR', 'KRW_SGL', 'KRW_SGR', 'KRW_SWRO', 'KRW_SWU',
                  'L', 'LAYER', 'LENGTH', 'MD', 'MDCON', 'ND', 'PARENT', 'PORTYPE', 'PPERF', 'PVMIN', 'RADB', 'RADBP',
                  'RADW', 'RADWP', 'RKHMIN', 'RPVMIN', 'SECT', 'SGL', 'SGR', 'SGRO', 'SGRW', 'SGTR', 'SGU', 'SKIN',
                  'SOTR', 'STAT', 'SWL', 'SWR', 'SWRO', 'SWU', 'TEMP', 'TIME', 'VAR', 'WELLMOD', 'WELLSPEC', 'WI', 'X',
                  'Y', 'ZONE']



"""The abstract base class for all wells."""
from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional, Sequence, Union

from ResSimpy.Completion import Completion
from ResSimpy.Enums.UnitsEnum import UnitSystem


@dataclass
class Well(ABC):
    __completions: list[Completion]
    __well_name: str
    __unit_system: UnitSystem

    def __init__(self, well_name: str, completions: list[Completion], unit_system: UnitSystem) -> None:
        self.__well_name = well_name
        self.__completions = completions
        self.__unit_system = unit_system

    @property
    def completions(self) -> list[Completion]:
        return self.__completions

    @property
    def well_name(self) -> str:
        return self.__well_name

    @property
    def unit_system(self) -> UnitSystem:
        return self.__unit_system

    @property
    def dates_of_completions(self) -> list[str]:
        """Returns a list of dates that the well was changed using a completion."""

        dates_changed: list[str] = []
        for completion in self.__completions:
            if completion.date not in dates_changed:
                dates_changed.append(completion.date)

        return dates_changed

    @property
    def perforations(self) -> Sequence[Completion]:
        """Returns a list of all of the perforations for the well."""
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @property
    def first_perforation(self) -> Optional[Completion]:
        """Returns the first perforation for the well."""
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @property
    def shutins(self) -> Sequence[Completion]:
        """Returns a list of all of the perforations for the well."""
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @property
    def last_shutin(self) -> Optional[Completion]:
        """Returns the first perforation for the well."""
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @property
    def printable_well_info(self) -> str:
        """Returns some printable well information in string format."""
        raise NotImplementedError("This method has not been implemented for this simulator yet")

    @property
    def completion_events(self) -> list[tuple[str, Union[int, tuple[float, float]]]]:
        """Returns a list of dates and values representing either the layer, or the depths of each perforation."""
        raise NotImplementedError("This method has not been implemented for this simulator yet")



"""Class representing a single Wellbore in a Nexus Network."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional

from ResSimpy.Wellbore import Wellbore


@dataclass(kw_only=True)
class NexusWellbore(Wellbore):
    """Class representing a single Wellbore in a Nexus Network.

    Attributes:
        date (str): string representation of the last defined date
        unit_system (UnitSystem): unit system enum
        name (str): Name of the well. (WELL)
        flowsect (int): Number of the flow section. (FLOWSECT)
        diameter (float): Diameter of the well. (DIAM)
        inner_diameter (float): Inner diameter of the well. (INNERDIAM)
        roughness (float): Roughness of the well. (ROUGHNESS)
        bore_type (str): Type of well. (TYPE)
        hyd_method (str): hydraulic method. (METHOD)
        temperature (float): Temperature of the well. (TEMP)
        elevation_profile (str): Elevation profile of the well. (ELEVPR)
        temperature_profile (str): Temperature profile of the well. (TEMPPR)
        heat_transfer_coeff (float): Heat transfer coefficient of the well. (HTC)
        pvt_method (int): Method number used for PVT. (IPVT)
        water_method (int): Method number used for water. (IWAT).
    """

    flowsect: Optional[int] = None
    bore_type: Optional[str] = None
    hyd_method: Optional[str] = None
    temperature: Optional[float] = None
    elevation_profile: Optional[str] = None
    temperature_profile: Optional[str] = None
    heat_transfer_coeff: Optional[float] = None
    pvt_method: Optional[int] = None
    water_method: Optional[int] = None

    def __init__(self, properties_dict: dict[str, None | int | str | float]) -> None:
        # call the init of the DataObjectMixin
        super(Wellbore, self).__init__({})
        for key, prop in properties_dict.items():
            self.__setattr__(key, prop)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        nexus_mapping = {
            'WELL': ('name', str),
            'FLOWSECT': ('flowsect', int),
            'DIAM': ('diameter', float),
            'INNERDIAM': ('inner_diameter', float),
            'ROUGHNESS': ('roughness', float),
            'TYPE': ('bore_type', str),
            'METHOD': ('hyd_method', str),
            'TEMP': ('temperature', float),
            'ELEVPR': ('elevation_profile', str),
            'TEMPPR': ('temperature_profile', str),
            'HTC': ('heat_transfer_coeff', float),
            'IPVT': ('pvt_method', int),
            'IWAT': ('water_method', int)
        }

        return nexus_mapping

    def __repr__(self) -> str:
        return super().__repr__()



"""Class to hold Nexus Rock properties."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.rock_keywords import ROCK_ALL_TABLE_KEYWORDS, ROCK_KEYWORDS_VALUE_FLOAT
from ResSimpy.Nexus.NexusKeywords.rock_keywords import ROCK_SINGLE_KEYWORDS, ROCK_KEYWORDS_VALUE_STR
from ResSimpy.Nexus.NexusKeywords.rock_keywords import ROCK_KEYWORDS, ROCK_REV_IRREV_OPTIONS
from ResSimpy.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import RockUnits
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits
from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusRockMethod(DynamicProperty):
    """Class to hold Nexus Rock properties.

    Attributes:
        file (NexusFile): Nexus rock properties file object
        input_number (int): Rock properties method number in Nexus fcs file
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific rock properties method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile
    properties: dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                dict[str, Union[float, pd.DataFrame]]]] = field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'CR': ('rock_compressibility', float),
            'KP': ('rock_permeability_compressibility', float),
            'PREF': ('reference_pressure', float),
            'P': ('pressure', float),
            'DP': ('delta_pressure', float),
        }
        return keywords

    @property
    def units(self) -> RockUnits:
        """Returns the attribute to unit map for the rock method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return RockUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with rock properties data, in Nexus file format."""
        printable_str = ''
        rock_dict = self.properties
        for key, value in rock_dict.items():
            if key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif isinstance(value, pd.DataFrame):
                printable_str += f'{key}\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n\n'
            elif isinstance(value, dict):
                printable_str += f"{key.replace('_',' ')}\n"
                for subkey in value.keys():
                    printable_str += f"SWINIT {subkey}\n"
                    df = value[subkey]
                    if isinstance(df, pd.DataFrame):
                        printable_str += df.to_string(na_rep='', index=False) + '\n'
                    printable_str += '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif value == '':
                printable_str += f'{key}\n'
            else:
                printable_str += f'{key} {value}\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus rock properties file contents and populate NexusRockMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize properties
        cmt_indices: list[int] = []
        wirct_indices_dict: dict[str, dict[str, list[int]]] = {}
        # Flag to tell when to start reading a table
        start_reading_table: bool = False
        swinit_key = ''
        # Dictionary of flags indicating which tables are being read
        table_being_read: dict[str, bool] = {}
        for table_name in ROCK_ALL_TABLE_KEYWORDS:
            table_being_read[table_name] = False

        line_indx = 0
        for line in file_as_list:

            # Find ROCK key-value pairs, such as PREF 2000 or CR 1e-6
            if [i for i in line.split() if i in ROCK_KEYWORDS_VALUE_FLOAT]:
                for key in ROCK_KEYWORDS_VALUE_FLOAT:
                    if nfo.check_token(key, line):
                        self.properties[key] = float(nfo.get_expected_token_value(key, line, file_as_list))
            # Find standalone rock property keywords, such as COMPR or KPMULT
            if [i for i in line.split() if i in ROCK_SINGLE_KEYWORDS]:
                for word in ROCK_SINGLE_KEYWORDS:
                    if nfo.check_token(word, line):
                        self.properties[word] = ''
            # Handle REVERSIBLE or IRREVERSIBLE keywords
            if [i for i in line.split() if i in ROCK_KEYWORDS_VALUE_STR]:
                for key in ROCK_KEYWORDS_VALUE_STR:
                    if nfo.check_token(key, line):
                        if nfo.get_token_value(key, line, file_as_list) in ROCK_REV_IRREV_OPTIONS:
                            self.properties[key] = nfo.get_expected_token_value(key, line, file_as_list)
                        else:
                            self.properties[key] = ''
            # Find starting index of rock compaction table
            if nfo.check_token('CMT', line):
                cmt_indices = [line_indx + 1, len(file_as_list)]
                table_being_read['CMT'] = True
                start_reading_table = True
                line_indx += 1
                continue
            # Find ending index of rock compaction table
            if start_reading_table and table_being_read['CMT']:
                for potential_endkeyword in ROCK_KEYWORDS:
                    if nfo.check_token(potential_endkeyword, line):
                        cmt_indices[1] = line_indx
                        start_reading_table = False
                        table_being_read['CMT'] = False
                        break
            # Find ending index of a water-induced rock compaction table
            if start_reading_table and table_being_read['WIRCT']:
                for potential_endkeyword in ROCK_KEYWORDS:
                    if nfo.check_token(potential_endkeyword, line):
                        wirct_indices_dict['WIRCT'][swinit_key][1] = line_indx
                        start_reading_table = False
                        table_being_read['WIRCT'] = False
                        break
            # Find starting index of a water-induced rock compaction table
            if nfo.check_token('WIRCT', line):
                wirct_indices_dict['WIRCT'] = {}
            if nfo.check_token('SWINIT', line):
                swinit_key = nfo.get_expected_token_value('SWINIT', line, file_as_list)
                wirct_indices_dict['WIRCT'][swinit_key] = [line_indx+1, len(file_as_list)]
                table_being_read['WIRCT'] = True
                start_reading_table = True

            line_indx += 1

        # Read in compaction table(s) if there are any
        if len(cmt_indices) > 0:
            self.properties['CMT'] = nfo.read_table_to_df(file_as_list[cmt_indices[0]:cmt_indices[1]])
        for key in wirct_indices_dict.keys():
            self.properties[key] = {}
            for subkey in wirct_indices_dict[key].keys():
                property_key = self.properties[key]
                if not isinstance(property_key, dict):
                    raise ValueError(f"Property is not a dictionary: {str(self.properties[key])}")
                property_key[subkey] = nfo.read_table_to_df(file_as_list[
                                                            wirct_indices_dict[key][subkey][0]:
                                                            wirct_indices_dict[key][subkey][1]])



"""Class for handling targets in the Nexus Network. This class is used to store and manipulate the targets in a
NexusNetwork. It is stored as an instance in the NexusNetwork class as "targets".
"""
from __future__ import annotations
from dataclasses import dataclass, field
from uuid import UUID
from typing import Sequence, Optional, TYPE_CHECKING

import pandas as pd

from ResSimpy.File import File
from ResSimpy.Nexus.nexus_add_new_object_to_file import AddObjectOperations
from ResSimpy.Nexus.nexus_collect_tables import collect_all_tables_to_objects
from ResSimpy.Nexus.DataModels.Network.NexusTarget import NexusTarget
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.nexus_modify_object_in_file import ModifyObjectOperations
from ResSimpy.Nexus.nexus_remove_object_from_file import RemoveObjectOperations
from ResSimpy.Targets import Targets
from ResSimpy.Utils.obj_to_dataframe import obj_to_dataframe

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass(kw_only=True)
class NexusTargets(Targets):
    """Class for handling targets in the Nexus Network. This class is used to store and manipulate the targets in a
    NexusNetwork. It is stored as an instance in the NexusNetwork class as "targets".
    """
    __targets: list[NexusTarget] = field(default_factory=list)

    def __init__(self, parent_network: NexusNetwork) -> None:
        self.__parent_network: NexusNetwork = parent_network
        self.__targets: list[NexusTarget] = []
        self.__add_object_operations = AddObjectOperations(None, self.table_header,
                                                           self.table_footer, self.__parent_network.model)
        self.__remove_object_operations = RemoveObjectOperations(self.__parent_network, self.table_header,
                                                                 self.table_footer)
        self.__modify_object_operations = ModifyObjectOperations(self)

    @property
    def table_header(self) -> str:
        """Start of the target definition table."""
        return 'TARGET'

    @property
    def table_footer(self) -> str:
        """End of the target definition table."""
        return 'ENDTARGET'

    def get_all(self) -> Sequence[NexusTarget]:
        """Returns a list of targets loaded from the simulator."""
        self.__parent_network.get_load_status()
        return self.__targets

    def get_by_name(self, target_name: str) -> Optional[NexusTarget]:
        """Returns a single target with the provided name loaded from the simulator.

        Args:
        ----
            target_name (str): name of the requested node

        Returns:
        -------
            NexusTarget: which has the same name as the requested target_name

        """
        targets_to_return = filter(lambda x: False if x.name is None else x.name.upper() == target_name.upper(),
                                   self.__targets)
        return next(targets_to_return, None)

    def get_df(self) -> pd.DataFrame:
        """Creates a dataframe representing all processed target data in a surface file
        Returns:
            DataFrame: of the properties of the targets through time with each row representing a target.
        """
        df_store = obj_to_dataframe(self.__targets)
        return df_store

    def get_overview(self) -> str:
        raise NotImplementedError('To be implemented')

    def load(self, surface_file: File, start_date: str, default_units: UnitSystem) -> None:
        """Calls load target and appends the list of discovered nodes into the NexusTargets object.

        Args:
            surface_file (File): NexusFile representation of the surface file.
            start_date (str): Starting date of the run
            default_units (UnitSystem): Units used in case not specified by surface file.

        Raises:
            TypeError: if the unit system found in the property check is not a valid enum UnitSystem.

        """
        new_targets = collect_all_tables_to_objects(surface_file, {'TARGET': NexusTarget},
                                                    start_date=start_date,
                                                    default_units=default_units)
        cons_list = new_targets.get('TARGET')
        if isinstance(cons_list, dict):
            raise ValueError('Incompatible data format for additional wells. Expected type "list" instead got "dict"')
        self._add_to_memory(cons_list)

    def _add_to_memory(self, additional_list: Optional[list[NexusTarget]]) -> None:
        """Extends the targets object by a list of targets provided to it.

        Args:
        ----
            additional_list (Sequence[NexusTarget]): list of nexus targets to add to the targets list.

        Returns:
        -------
            None
        """
        if additional_list is None:
            return
        self.__targets.extend(additional_list)

    def remove(self, target_to_remove: dict[str, None | str | float | int] | UUID) -> None:
        """Remove a node from the network based on the properties matching a dictionary or id.

        Args:
            target_to_remove (UUID | dict[str, None | str | float | int]): UUID of the node to remove or a dictionary \
            with sufficient matching parameters to uniquely identify a node

        """
        self.__remove_object_operations.remove_object_from_network_main(
            target_to_remove, self._network_element_name, self.__targets)

    def add(self, target_to_add: dict[str, None | str | float | int]) -> None:
        """Adds a target to a network, taking a dictionary with properties for the new node.

        Args:
            target_to_add (dict[str, None | str | float | int]):
            dictionary taking all the properties for the new target.
            Requires date and a target name.
        """
        new_object = self.__add_object_operations.add_network_obj(target_to_add, NexusTarget,  self.__parent_network)
        self._add_to_memory([new_object])

    def modify(self, target_to_modify: dict[str, None | str | float | int],
               new_properties: dict[str, None | str | float | int]) -> None:
        """Modifies an existing node based on a matching dictionary of properties (partial matches allowed if precisely
         1 matching node is found). Updates the properties with properties in the new_properties dictionary.

        Args:
            target_to_modify (dict[str, None | str | float | int]): dictionary containing attributes to match in the
            existing node set.
            new_properties (dict[str, None | str | float | int]): properties to switch to in the new node
        """
        self.__parent_network.get_load_status()

        self.__modify_object_operations.modify_network_object(target_to_modify, new_properties,
                                                              self.__parent_network)



from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Enums.UnitsEnum import SUnits, TemperatureUnits, UnitSystem
from ResSimpy.Nexus.NexusKeywords.gaslift_keywords import GL_ARRAY_KEYWORDS, GASLIFT_KEYWORDS, GL_TABLE_HEADER_COLS
from ResSimpy.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import HydraulicsUnits

from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusGasliftMethod(DynamicProperty):
    """Class to hold Nexus gaslift properties.

    Attributes:
        file (NexusFile): Nexus gaslift properties file object
        input_number (int): Gaslift properties method number in Nexus fcs file
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific gaslift properties method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile
    properties: dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                     dict[str, Union[float, pd.DataFrame]]]] = field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'QOIL': ('surface_oil_rate', float),
            'QLIQ': ('surface_liquid_rate', float),
            'PRESSURE': ('pressure', float),
            'WCUT': ('watercut', float),
            'GLR': ('gas_liquid_ratio', float),
            'GOR': ('gas_oil_ratio', float),
        }
        return keywords

    @property
    def units(self) -> HydraulicsUnits:
        """Returns the attribute to unit map for the Gaslift method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return HydraulicsUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with gaslift data in Nexus file format."""
        printable_str = ''
        gl_dict = self.properties
        for key, value in gl_dict.items():
            if isinstance(value, pd.DataFrame):
                printable_str += value.to_string(na_rep='', index=False) + '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem | TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif value == '':
                printable_str += f'{key}\n'
            else:
                printable_str += f'{key} {value}\n'
        printable_str += '\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus gaslift file contents and populate the NexusGasliftMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers to use to record properties as we iterate through aquifer file contents
        # Dictionary to record start and ending indices for tables
        gl_table_indices: dict[str, list[int]] = {}
        # Dictionary of flags indicating which tables are being read
        start_reading_table = False

        line_indx = 0
        for line in file_as_list:

            # Find arrays of parameters, e.g., QOIL 1.0 10. 100., or WCUT 0.0 0.1 0.2
            potential_keyword = nfo.check_list_tokens(GL_ARRAY_KEYWORDS, line)
            if potential_keyword is not None:
                line_elems = line.split('!')[0].split()
                keyword_index = line_elems.index(potential_keyword)
                self.properties[potential_keyword] = ' '.join(line_elems[keyword_index+1:])

            # Find ending index of gaslift table
            if start_reading_table:
                for potential_endkeyword in GASLIFT_KEYWORDS:
                    if nfo.check_token(potential_endkeyword, line):
                        gl_table_indices['GL_TABLE'][1] = line_indx
                        start_reading_table = False
                        break
            # Find starting index of gaslift table
            for header_keyword in GL_TABLE_HEADER_COLS:
                if nfo.check_token(header_keyword, line):
                    gl_table_indices['GL_TABLE'] = [line_indx, len(file_as_list)]
                    start_reading_table = True

            line_indx += 1

        # Read in gaslift tables
        for key in gl_table_indices.keys():
            self.properties[key] = nfo.read_table_to_df(file_as_list[
                gl_table_indices[key][0]:gl_table_indices[key][1]])



from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataObjectMixin import DataObjectMixin
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Units.AttributeMappings.NetworkUnitMapping import NetworkUnits


@dataclass(repr=False)
class Wellhead(DataObjectMixin, ABC):
    well: Optional[str] = None
    name: Optional[str] = None
    date: Optional[str] = None
    unit_system: Optional[UnitSystem] = None
    wellhead_type: Optional[str] = None
    depth: Optional[float] = None
    x_pos: Optional[float] = None
    y_pos: Optional[float] = None

    @property
    def units(self) -> NetworkUnits:
        """Returns the attribute to unit map for the WellConnection."""
        return NetworkUnits(self.unit_system)



"""A class to modify an existing object in a file with a new set of properties."""
from __future__ import annotations

from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


class ModifyObjectOperations:
    def __init__(self, object_to_modify: Any) -> None:
        self.object_to_modify = object_to_modify

    def modify_network_object(self, object_to_modify: dict[str, None | str | float | int],
                              new_properties: dict[str, None | str | float | int], network: NexusNetwork) -> None:
        """Modifies an existing object based on a matching dictionary of properties (partial matches allowed if
        precisely 1 matching object is found).
        Updates the properties with properties in the new_properties dictionary.
        Applies primarily to network based objects.

        Args:
            object_to_modify (dict[str, None | str | float | int]): dictionary containing attributes to match in the
            existing object set. Requires an implemented add, remove
            new_properties (dict[str, None | str | float | int]): properties to switch to in the new object
        """
        # TODO apply this to more of the network attributes through the Base Class
        network_attribute_name = self.object_to_modify._network_element_name
        name = object_to_modify.get('name', None)
        if name is None:
            raise ValueError(f'Name is required for modifying {network_attribute_name}, '
                             f'instead got {name}')
        name = str(name)
        network_element = network.find_network_element_with_dict(name, object_to_modify,
                                                                 network_attribute_name)
        existing_properties = network_element.to_dict(include_nones=False)
        # do the union of the two dicts
        existing_properties.update(new_properties)

        self.object_to_modify.remove(object_to_modify)
        self.object_to_modify.add(existing_properties)



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional


@dataclass
class VariableEntry:
    modifier: Optional[str] = None
    value: Optional[str] = None


@dataclass(kw_only=True)
class Grid(ABC):
    # Grid dimensions
    _range_x: Optional[int]
    _range_y: Optional[int]
    _range_z: Optional[int]

    _netgrs: VariableEntry
    _porosity: VariableEntry
    _sw: VariableEntry
    _kx: VariableEntry
    _ky: VariableEntry
    _kz: VariableEntry

    def __init__(self) -> None:
        self._netgrs = VariableEntry()
        self._porosity = VariableEntry()
        self._sw = VariableEntry()
        self._kx = VariableEntry()
        self._ky = VariableEntry()
        self._kz = VariableEntry()
        # Grid dimensions
        self._range_x: Optional[int] = None
        self._range_y: Optional[int] = None
        self._range_z: Optional[int] = None

    @property
    def range_x(self):
        self.load_grid_properties_if_not_loaded()
        return self._range_x

    @property
    def range_y(self):
        self.load_grid_properties_if_not_loaded()
        return self._range_y

    @property
    def range_z(self):
        self.load_grid_properties_if_not_loaded()
        return self._range_z

    @property
    def netgrs(self):
        self.load_grid_properties_if_not_loaded()
        return self._netgrs

    @property
    def porosity(self):
        self.load_grid_properties_if_not_loaded()
        return self._porosity

    @property
    def sw(self):
        self.load_grid_properties_if_not_loaded()
        return self._sw

    @property
    def kx(self):
        self.load_grid_properties_if_not_loaded()
        return self._kx

    @property
    def ky(self):
        self.load_grid_properties_if_not_loaded()
        return self._ky

    @property
    def kz(self):
        self.load_grid_properties_if_not_loaded()
        return self._kz

    @abstractmethod
    def load_structured_grid_file(self, structure_grid_file, lazy_loading) -> Grid:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def load_grid_properties_if_not_loaded(self) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def to_dict(self) -> dict[str, Optional[int] | VariableEntry]:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def update_properties_from_dict(self, data: dict[str, int | VariableEntry]) -> None:
        raise NotImplementedError("Implement this in the derived class")



"""A class built for removing an object from a file and from the list of objects."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Literal, TypeVar, Optional
from uuid import UUID

from ResSimpy.DataObjectMixin import DataObjectMixin
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
T = TypeVar('T', bound=DataObjectMixin)

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


class RemoveObjectOperations:
    def __init__(self, network: Optional[NexusNetwork], table_header: str, table_footer: str) -> None:
        self.table_header = table_header
        self.table_footer = table_footer
        self.__network = network

    @staticmethod
    def remove_object_from_memory_by_id(list_obj: list[T], id_to_remove: UUID) -> tuple[Any, list[Any]]:
        """Directly removes an object from a list of objects based on the id attribute of that object."""
        if len(list_obj) == 0:
            raise ValueError('Tried to remove object from empty list. Cannot remove object.')

        if not hasattr(list_obj[0], 'id'):
            raise AttributeError(f'Objects provided in {list_obj} has no attribute id.')

        index_to_remove = [x.id for x in list_obj].index(id_to_remove)
        object_removed = list_obj.pop(index_to_remove)
        return object_removed, list_obj

    def check_for_empty_table(self, file: NexusFile, line_numbers_in_file_to_remove: list[int], obj_id: UUID) \
            -> list[int]:
        """Identifies the lines needed to be removed if the table is empty.

        Args:
            file (NexusFile): file to check
            line_numbers_in_file_to_remove (list[int]): list of line indices for the object being removed
            obj_id (UUID): id of the object being removed

        Returns:
            A list of integers with the lines to remove from the file if the resulting table is empty after the lines\
             associated with the removed object is removed
        """
        first_obj_index = line_numbers_in_file_to_remove[0]
        last_obj_index = line_numbers_in_file_to_remove[-1]
        additional_indices_to_remove = []
        remove_table = True
        # get all the indices for the tables:
        file_content = file.get_flat_list_str_file
        start_node_keyword_index_to_remove = max([i for i, x in enumerate(file_content) if self.table_header in x and
                                                  i < first_obj_index])
        end_node_keyword_index_to_remove = min([i for i, x in enumerate(file_content) if self.table_footer in x and
                                                i > last_obj_index])
        # check there are any nodes left in the specified table
        if file.object_locations is None:
            raise ValueError(f'No object locations specified, cannot find id: {obj_id} in {file.object_locations}')

        for obj_uuid, line_locations_list in file.object_locations.items():
            if obj_uuid == obj_id:
                # ignore the uuid's for the node that we want to remove
                continue
            for value in line_locations_list:
                # if we find an object in the middle of the table then don't remove it
                if start_node_keyword_index_to_remove <= value <= end_node_keyword_index_to_remove:
                    remove_table = False
                    break
        if remove_table:
            additional_indices_to_remove = list(range(start_node_keyword_index_to_remove,
                                                      end_node_keyword_index_to_remove + 1))
        return additional_indices_to_remove

    def remove_lines_from_file(self, line_numbers_in_file_to_remove: list[int], file: NexusFile, obj_id: UUID) -> None:
        """Removes lines from the file content and removes any references to the object in the line locations in the \
        file.

        Args:
            line_numbers_in_file_to_remove (list[int]): line indices to be removed. Relative to the flattened file.
            file (NexusFile): NexusFile to remove the lines from. Lines nested in files will be resolved correctly.
            obj_id (UUID): id of the object being removed
        """
        # get unique line numbers + sort them in descending order
        line_numbers_in_file_to_remove = list(set(line_numbers_in_file_to_remove))
        line_numbers_in_file_to_remove.sort(reverse=True)
        # remove the lines
        for index, line_in_file in enumerate(line_numbers_in_file_to_remove):
            if index == 0:
                file.remove_from_file_as_list(line_in_file, [obj_id])
            else:
                file.remove_from_file_as_list(line_in_file)

    def remove_object_by_id(self, file: NexusFile, obj_id: UUID, obj_list: list[T]) -> None:
        """Remove an object from a file and from the list of stored objects.

        Args:
            file (NexusFile): file to remove the lines associated with the obj from.
            obj_id (UUID): id of the object being removed.
            obj_list (list[Any]): list of corresponding objects.
        """
        # remove from memory
        self.remove_object_from_memory_by_id(obj_list, id_to_remove=obj_id)
        # remove from file
        line_numbers_in_file_to_remove = file.get_object_locations_for_id(obj_id)
        # get table_header and footers
        remove_empty_table_indices = self.check_for_empty_table(
            file, line_numbers_in_file_to_remove, obj_id)
        # remove the table if there aren't any more remaining
        line_numbers_in_file_to_remove.extend(remove_empty_table_indices)
        self.remove_lines_from_file(line_numbers_in_file_to_remove, file, obj_id)

    def remove_object_from_network_main(self, obj_to_remove: dict[str, None | str | float | int] | UUID,
                                        network_element_name: Literal['nodes', 'connections', 'well_connections',
                                                                      'wellheads', 'wellbores', 'constraints',
                                                                      'targets'],
                                        existing_objects: list[T]) -> None:
        """Removes object from file and from the list of objects based on matching a set of attributes provided in a
        dictionary or a unique id.

        Args:
            obj_to_remove (dict[str, None | str | float | int] | UUID): dictionary of properties for identifying which
            object to remove. Can also take an object id.
            network_element_name (str): One of: 'nodes', 'connections', 'well_connections', 'wellheads', 'wellbores',
            'constraints'. Identifies the attribute name of the element inside the network
            existing_objects (list[T]): list of all existing network objects for the network element. e.g. self.__nodes
        """
        if self.__network is None:
            raise ValueError('No network provided for removing objects from.'
                             'Cannot remove an object from something that does not exist')
        self.__network.get_load_status()
        network_file = self.__network.get_network_file()
        if isinstance(obj_to_remove, dict):
            name = obj_to_remove.get('name', None)
            if name is None:
                raise ValueError(f'Require node name to remove the node instead got {name=}')
            name = str(name)
            obj_in_network = self.__network.find_network_element_with_dict(name, obj_to_remove,
                                                                           network_element_name)
            obj_id = obj_in_network.id
        else:
            obj_id = obj_to_remove
        self.remove_object_by_id(network_file, obj_id, existing_objects)



# TODO: Not a complete list. Check manual for missing keywords.
# TODO: Delete the keywords that are not tokens (i.e. that are only values in a table)

INTEGER_ARRAYS = ['ICOARS', 'IEQUIL', 'IPVT', 'IWATER', 'IALPHAF', 'IPOLYMER', 'IRELPM', 'IROCK', 'IREGION',
                  'IADSORPTION', 'ITRAN', 'ITRACER', 'DEADCELL', 'LIVECELL', 'IGRID', 'ISECTOR']

WORK_ARRAYS = ['WORKA1', 'WORKA2', 'WORKA3', 'WORKA4', 'WORKA5', 'WORKA6', 'WORKA7', 'WORKA8', 'WORKA9']

ROCK_ARRAYS = ['SWL', 'SWR', 'SGL', 'SGR', 'SWRO', 'SWRO_LS', 'SGRO', 'SGRW', 'KRW_SWRO', 'KRWS_LS', 'KRW_SWU',
               'KRG_SGRO', 'KRG_SGU', 'KRO_SWL', 'KRO_SGL', 'KRO_SGR', 'KRW_SGL', 'KRW_SGR', 'KRG_SGRW', 'SGTR',
               'SOTR', 'SWLPC', 'SGLPC', 'PCW_SWL', 'PCG_SGU']

USER_INIT_ARRAYS = ['SW', 'SG', 'PRESSURE', 'TEMPERATURE', 'CHLORIDE', 'CALCIUM', 'SALINITY', 'API',
                    'WORKA1', 'WORKA2', 'WORKA3', 'WORKA4', 'WORKA5', 'WORKA6', 'WORKA7', 'WORKA8', 'WORKA9']

MULTIPLIER_ARRAYS = ['TMX', 'TMY', 'TMZ', 'MULTBV']

PROPERTY_ARRAYS = ['POROSITY', 'POR', 'POROSITY', 'PV', 'KX', 'KY', 'KZ', 'COMPR', 'CR', 'NETGRS', 'KWX',
                   'KWY', 'KWZ', 'PEMDMAT', 'PEMGMAT', 'PEMKMAT', 'BW_SAL', 'BW_T']

GRID_GEOMETRY_ARRAYS = ['CORP', 'EIGHT', 'DX', 'DY', 'DZ', 'DXB', 'DYB', 'DZB', 'MDEPTH', 'DEPTH', 'DBZNET']

GRID_OPERATION_KEYWORDS = ['ADD', 'SUB', 'DIV', 'MULT', 'EQ']

GRID_ARRAY_FORMAT_KEYWORDS = ['VALUE', 'XVAR', 'YVAR', 'ZVAR']

GRID_ARRAY_KEYWORDS = INTEGER_ARRAYS + WORK_ARRAYS + ROCK_ARRAYS + USER_INIT_ARRAYS + MULTIPLIER_ARRAYS + \
                      PROPERTY_ARRAYS + GRID_GEOMETRY_ARRAYS

STRUCTURED_GRID_KEYWORDS = ['ADD', 'ALL', 'ANALYT', 'ARRAYS', 'B', 'BLOCKS', 'C', 'CARTREF', 'COARSEN', 'COMPR', 'CON',
                            'CONCENTRATION', 'CORNER', 'CORP', 'CORTOL', 'DEADCELL', 'DECOMP', 'DEPTH', 'DIP', 'DIV',
                            'DX', 'DY', 'DZ', 'END', 'ENDAQ', 'ENDDEC', 'ENDREF', 'EXP', 'FNAME', 'FRAC', 'FTRANS',
                            'FUNCTION', 'GE', 'GRID', 'ID', 'IEQUIL', 'IINF', 'INCLUDE', 'INFLUX', 'IPVT', 'IREGION',
                            'IRELPM', 'IROCK', 'ITRAN', 'IWATER', 'JINF', 'KINF', 'KRG_SGRW', 'KRO_SWL', 'KRW_SWRO',
                            'KRW_SWU', 'KX', 'KXEFF', 'KY', 'KYEFF', 'KZ', 'KZEFF', 'LAYER', 'LE', 'LGR', 'LIVECELL',
                            'LX', 'LY', 'LZ', 'MAPBINARY', 'MAPOUT', 'MAPVDB', 'MDEPTH', 'MIN', 'MINUS', 'MOD', 'MODX',
                            'MODY', 'MODZ', 'MULT', 'MULTBV', 'MULTFL', 'MULTIR', 'NETGRS', 'NEWTRAN', 'NOLIST', 'NONE',
                            'NONSTD', 'NX', 'NY', 'NZ', 'OMIT', 'OUTPUT', 'OVER', 'PINCHOUT', 'POR', 'POROSITY',
                            'PRINT', 'PV', 'PVMULT', 'RANGE', 'RIGHTHANDED', 'ROOT', 'SALINITY', 'SG', 'SGL', 'SGR',
                            'SGRO', 'SGRW', 'SGU', 'SINF', 'STD', 'SW', 'SWL', 'SWR', 'SWRO', 'SWRO_LS', 'SWU', 'TMX',
                            'TMY', 'TMZ', 'TOLPV', 'TX', 'TY', 'TZ', 'V98', 'VALUE', 'WATER', 'WINDOW', 'WORKA1', 'X',
                            'XREG', 'Z', 'ZVAR']



from __future__ import annotations
from typing import Optional
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from datetime import datetime, timedelta


class ISODateTime(datetime):
    """A class representing an extension of datetime class,  returns back date in ISO datetime format."""

    def __repr__(self) -> str:
        """Return the object representation, but formatted in ISO format."""
        basic_string = super().__repr__()
        iso_string = basic_string.replace(' ', 'T')
        return iso_string

    def __str__(self) -> str:
        """Return the string representation, but formatted in ISO format."""
        basic_string = super().__str__()
        iso_string = basic_string.replace(' ', 'T')
        return iso_string

    @staticmethod
    def isfloat(no_of_days: str) -> bool:
        if no_of_days is not None:
            try:
                float(no_of_days)
                return True
            except ValueError:
                return False
        else:
            return False

    @classmethod
    def convert_to_iso(cls, date: str, date_format: DateFormat, start_date: Optional[str] = None) -> ISODateTime:
        """Converts an ordinary date to an ISODate format.

        Args:
            date (str): The date as it is written in the model file.
            date_format (DateFormat): The date format to use to convert the date.
            start_date (Optional[str]): The start date of the model (required if the date is a number of days from the
                start).

        Raises:
            ValueError: if the provided parameters cannot produce a valid date.

        Returns:
        ISODateTime: an initialised ISODateTime class instance for the date provided.
        """
        converted_date = None

        if date_format is None:
            raise ValueError('Please provide a date format')

        if ISODateTime.isfloat(date) and start_date is None:
            raise ValueError('Please provide start date when date is numeric')
        elif ISODateTime.isfloat(date) and start_date is not None:
            if date_format == DateFormat.DD_MM_YYYY:
                converted_date = ISODateTime.strptime(start_date, '%d/%m/%Y') + timedelta(days=float(date))
            elif date_format == DateFormat.MM_DD_YYYY:
                converted_date = ISODateTime.strptime(start_date, '%m/%d/%Y') + timedelta(days=float(date))

        elif date_format == DateFormat.DD_MM_YYYY:
            converted_date = ISODateTime.strptime(date, '%d/%m/%Y')

        elif date_format == DateFormat.MM_DD_YYYY:
            converted_date = ISODateTime.strptime(date, '%m/%d/%Y')

        elif date_format == DateFormat.DD_MMM_YYYY:
            converted_date = ISODateTime.strptime(date, '%d %b %Y')

        if converted_date is None:
            raise ValueError('Invalid date format or missing start_date.')

        return converted_date

    @classmethod
    def datetime_to_iso(cls, date: datetime, datetime_format: str = '%Y-%m-%d') -> ISODateTime:
        """Converts datetime object to ISODateTime object."""
        return ISODateTime.strptime(str(date), datetime_format)



"""The base class for all Well Completions."""
from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataObjectMixin import DataObjectMixin
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.ISODateTime import ISODateTime
from ResSimpy.Nexus.NexusEnums import DateFormatEnum
from ResSimpy.Units.AttributeMappings.CompletionUnitMapping import CompletionUnits


@dataclass(kw_only=True, repr=False)
class Completion(DataObjectMixin, ABC):
    """A class representing well completions.

    IMPORTANT: if modifying this class, make sure to update the relevant tests in test_load_wells, as well as updating
    the constructor calls in the derived classes.

    Args:
    ----
        date (str): The starting date of the completion. Applies until changed.
        i (Optional[int]): The structured grid cell location in the x direction. 'IW' in Nexus
        j (Optional[int]): The structured grid cell location in the y direction. 'JW' in Nexus
        k (Optional[int]): The structured grid cell location in the z direction. 'L' in Nexus
        skin (Optional[float]): The skin value for the completion. 'SKIN' in Nexus
        depth (Optional[float]): The depth of the completion. 'DEPTH' in Nexus
        well_radius (Optional[float]): The well radius. 'RADW' in Nexus
        x (Optional[float]): The x location of the well in distance units/coordinates. 'X' in Nexus
        y (Optional[float]): The y location of the well in distance units/coordinates. 'Y' in Nexus
        angle_a (Optional[float]): the angle relative to the local I axis. 'ANGLA' in Nexus.
        angle_v (Optional[float]): the angle relative to the true vertical axis (global Z axis). 'ANGLV' in Nexus
        grid (Optional[str]): the grid name to which the completion data applies. 'GRID' in Nexus
        depth_to_top (Optional[float]): subsea depth to the top of a completion interval. 'DTOP' in Nexus
        depth_to_bot (Optional[float]): subsea depth to the bottom of the completion interval. 'DBOT' in Nexus
        perm_thickness_ovr (Optional[float]): permeability thickness override value to use for the completion interval.\
            'KH' in Nexus.
        dfactor (Optional[float]): non-darcy factor to use for rate dependent skin calculations. 'D' in Nexus
        rel_perm_method (Optional[int]): rel perm method to use for the completion. 'IRELPM' in Nexus
        status (Optional[str]): the status of the layer, can be 'ON' or 'OFF'
        peaceman_well_block_radius (Optional[float]): The pressure equivalent radius of the grid block
    """

    __date: str
    __i: Optional[int] = None
    __j: Optional[int] = None
    __k: Optional[int] = None
    __skin: Optional[float] = None
    __depth: Optional[float] = None
    __well_radius: Optional[float] = None
    __x: Optional[float] = None
    __y: Optional[float] = None
    __angle_a: Optional[float] = None
    __angle_v: Optional[float] = None
    __grid: Optional[str] = None
    __depth_to_top: Optional[float] = None
    __depth_to_bottom: Optional[float] = None
    __perm_thickness_ovr: Optional[float] = None
    __dfactor: Optional[float] = None
    __rel_perm_method: Optional[int] = None
    __status: Optional[str] = None
    __iso_date: Optional[ISODateTime] = None
    _date_format: Optional[DateFormatEnum.DateFormat] = None
    __start_date: Optional[str] = None
    __unit_system: Optional[UnitSystem] = None
    __peaceman_well_block_radius: Optional[float] = None

    def __init__(self, date: str, i: Optional[int] = None, j: Optional[int] = None, k: Optional[int] = None,
                 skin: Optional[float] = None, depth: Optional[float] = None, well_radius: Optional[float] = None,
                 x: Optional[float] = None, y: Optional[float] = None, angle_a: Optional[float] = None,
                 angle_v: Optional[float] = None, grid: Optional[str] = None, depth_to_top: Optional[float] = None,
                 depth_to_bottom: Optional[float] = None, perm_thickness_ovr: Optional[float] = None,
                 dfactor: Optional[float] = None, rel_perm_method: Optional[int] = None,
                 status: Optional[str] = None, date_format: Optional[DateFormatEnum.DateFormat] = None,
                 peaceman_well_block_radius: Optional[float] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None) -> None:
        super().__init__({})
        self._date_format = date_format
        self.__well_radius = well_radius
        self.__date = date
        self.__i = i
        self.__j = j
        self.__k = k
        self.__skin = skin
        self.__depth = depth
        self.__x = x
        self.__y = y
        self.__angle_a = angle_a
        self.__angle_v = angle_v
        self.__grid = grid
        self.__depth_to_top = depth_to_top
        self.__depth_to_bottom = depth_to_bottom
        self.__perm_thickness_ovr = perm_thickness_ovr
        self.__dfactor = dfactor
        self.__rel_perm_method = rel_perm_method
        self.__status = status
        self.__start_date = start_date
        self.__iso_date = self.set_iso_date()
        self.__unit_system = unit_system
        self.__peaceman_well_block_radius = peaceman_well_block_radius

    @property
    def well_radius(self):
        return self.__well_radius

    @property
    def date(self):
        return self.__date

    @property
    def iso_date(self):
        return self.__iso_date

    @property
    def i(self):
        return self.__i

    @property
    def j(self):
        return self.__j

    @property
    def k(self):
        return self.__k

    @property
    def skin(self):
        return self.__skin

    @property
    def depth(self):
        return self.__depth

    @property
    def x(self):
        return self.__x

    @property
    def y(self):
        return self.__y

    @property
    def angle_a(self):
        return self.__angle_a

    @property
    def angle_v(self):
        return self.__angle_v

    @property
    def grid(self):
        return self.__grid

    @property
    def depth_to_top(self):
        return self.__depth_to_top

    @property
    def depth_to_bottom(self):
        return self.__depth_to_bottom

    @property
    def perm_thickness_ovr(self):
        return self.__perm_thickness_ovr

    @property
    def dfactor(self):
        return self.__dfactor

    @property
    def rel_perm_method(self):
        return self.__rel_perm_method

    @property
    def status(self):
        return self.__status

    @property
    def peaceman_well_block_radius(self) -> Optional[float]:
        """The Peaceman Well Block Radius for the completion."""
        return self.__peaceman_well_block_radius

    @property
    def date_format(self):
        return self._date_format

    @property
    def start_date(self):
        return self.__start_date

    def set_iso_date(self) -> ISODateTime:
        return ISODateTime.convert_to_iso(self.date, self.date_format, self.start_date)

    @property
    def unit_system(self):
        return self.__unit_system

    @property
    def units(self) -> CompletionUnits:
        return CompletionUnits(self.unit_system)



"""Functions for collecting tables from a Nexus file to populate Nexus objects."""
from __future__ import annotations

from typing import Any, Optional

from ResSimpy.File import File
from ResSimpy.Nexus.DataModels.Network.NexusConstraint import NexusConstraint
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.nexus_constraint_operations import load_inline_constraints
from ResSimpy.Nexus.nexus_file_operations import check_property_in_line, check_token, get_expected_token_value, \
    check_list_tokens, load_table_to_objects


def collect_all_tables_to_objects(nexus_file: File, table_object_map: dict[str, Any], start_date: Optional[str],
                                  default_units: Optional[UnitSystem]) -> \
        dict[str, list[Any] | dict[str, list[NexusConstraint]]]:
    """Loads all tables from a given file.

    Args:
    ----
    nexus_file (File): NexusFile representation of the file.
    table_object_map (dict[str, Storage_Object]): dictionary containing the name of the table as keys and \
                the object type to store the data from each row into. Require objects to have a get_keyword_mapping \
                function
    model: (NexusSimulator): main simulator object

    Raises:
    ------
    TypeError: if the unit system found in the property check is not a valid enum UnitSystem.

    Returns:
    -------
    dict[str, list[Storage_Object]]: a dictionary of lists of arbitrary objects populated \
                with properties from the file provided, keyed with the NexusTable name associated with table_object_map.
    """
    current_date = start_date
    nexus_object_results: dict[str, list[Any] | dict[str, list[NexusConstraint]]] = {x: [] for x in table_object_map}
    nexus_constraints: dict[str, list[NexusConstraint]] = {}
    nexus_object_results['CONSTRAINTS'] = nexus_constraints
    file_as_list: list[str] = nexus_file.get_flat_list_str_file
    table_start: int = -1
    table_end: int = -1
    property_dict: dict = {}
    token_found: Optional[str] = None
    network_names: list[str] = []
    for index, line in enumerate(file_as_list):
        # check for changes in unit system
        check_property_in_line(line, property_dict, file_as_list)
        unit_system = property_dict.get('UNIT_SYSTEM', default_units)
        if not isinstance(unit_system, UnitSystem):
            raise TypeError(f"Value found for {unit_system=} of type {type(unit_system)} \
                                not compatible, expected type UnitSystem Enum")
        if check_token('TIME', line):
            current_date = get_expected_token_value(
                token='TIME', token_line=line, file_list=file_as_list,
                custom_message=f"Cannot find the date associated with the TIME card in {line=} at line number {index}")
            continue
        if table_start < 0:
            token_found = check_list_tokens(list(table_object_map.keys()), line)
            if token_found is None or check_token('WELLCONTROL', line):
                continue
            # if a token is found get the starting index of the table
            table_start = index + 1
        if token_found is None:
            continue
        token_found = token_found.upper()

        if table_start > 0 and check_token('END' + token_found, line):
            table_end = index
        # if we have a complete table to read in start reading it into objects
        if 0 < table_start <= table_end:
            # cover for the case where we aren't currently reading in the table to an object.
            # if no object is provided by the map for the token found then skip the keyword and reset the tracking vars
            if table_object_map[token_found] is None:
                table_start = -1
                table_end = -1
                token_found = None
                continue
            list_objects = None
            property_map = table_object_map[token_found].get_keyword_mapping()
            if token_found == 'CONSTRAINTS':
                load_inline_constraints(file_as_list=file_as_list[table_start:table_end],
                                        constraint=table_object_map[token_found],
                                        current_date=current_date,
                                        unit_system=unit_system,
                                        property_map=property_map,
                                        existing_constraints=nexus_constraints,
                                        nexus_file=nexus_file,
                                        start_line_index=table_start,
                                        network_names=network_names,
                                        )

            elif token_found == 'QMULT' or token_found == 'CONSTRAINT':
                list_objects = load_table_to_objects(file_as_list=file_as_list[table_start:table_end],
                                                     row_object=table_object_map[token_found],
                                                     property_map=property_map,
                                                     current_date=current_date,
                                                     unit_system=unit_system,
                                                     nexus_obj_dict=nexus_constraints,
                                                     preserve_previous_object_attributes=True)

            else:
                list_objects = load_table_to_objects(file_as_list=file_as_list[table_start:table_end],
                                                     row_object=table_object_map[token_found],
                                                     property_map=property_map,
                                                     current_date=current_date,
                                                     unit_system=unit_system)

            # store objects found into right dictionary
            list_of_token_obj = nexus_object_results[token_found]
            # This statement ensures that CONSTRAINT that are found in tables are actually added to the dictionary
            # under the same key as constraints to preserve their order
            if (token_found == 'CONSTRAINT' or token_found == 'QMULT') and list_objects is not None:
                for constraint, id_index in list_objects:
                    if isinstance(constraint, NexusConstraint):
                        obj_id = constraint.id
                        well_name = constraint.name
                    else:
                        obj_id = constraint
                        well_name = None
                    correct_line_index = id_index + table_start
                    nexus_file.add_object_locations(obj_id, [correct_line_index])
                    if well_name is None:
                        continue
                    if nexus_constraints.get(well_name, None) is not None:
                        nexus_constraints[well_name].append(constraint)
                    else:
                        nexus_constraints[well_name] = [constraint]
            elif list_objects is not None and isinstance(list_of_token_obj, list):
                list_of_token_obj.extend([x[0] for x in list_objects])
                # add the names from the nodes into the network names for wildcards
                network_names.extend([x.name for x in list_of_token_obj])
                for new_object, id_index in list_objects:
                    correct_line_index = id_index + table_start
                    # temporary try statement until all objects have an id property
                    try:
                        obj_id = new_object.id
                        nexus_file.add_object_locations(obj_id, [correct_line_index])
                    except AttributeError:
                        pass
            else:
                list_of_token_obj = nexus_constraints
            # reset indices for further tables
            table_start = -1
            table_end = -1
            token_found = None
    return nexus_object_results
